/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;

import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EngineRT;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactoryDependencies;
import org.bluestemsoftware.specification.eoa.ext.engine.EngineFactory;

/**
 * Loads classes and resources required by an <code>Application</code> at runtime, i.e. as
 * modeled by {@link EngineRT}. The delegation sequence is as follows:
 * <p>
 * First, the parent classloader, which SHOULD be sun.misc.Launcher$AppClassLoader and which
 * SHOULD contain all {@link ProvidedDependency} artifacts is queried.
 * <p>
 * Second, all artifacts scoped to this <code>ClassLoader</code>, if any, are searched.
 * <p>
 * Third, artifacts scoped to deployment containing {@link EngineFactory} that 'instantiated'
 * the application are searched.
 * <p>
 * Fourth, artifacts scoped to the {@link ComponentDeployment} from which runtime application
 * definition, i.e. as modeled by {@link ApplicationRT}, was loaded are searched.
 * <p>
 * Lastly, artifacts scoped to the system's {@link SharedClassLoader} are searched.
 * 
 * @see ExtensionFactoryDependencies
 */
public class ApplicationClassLoader extends AbstractClassLoader {

    private DeploymentClassLoader factoryLoader;
    private DeploymentClassLoader componentLoader;

    /**
     * Preferred constructor.
     * 
     * @param id
     *        identifier which SHOULD be unique within context of JVM
     * @param urls
     *        the URLs from which to load classes and resources, i.e. artifacts visible only to
     *        this classloader.
     * @param factoryLoader
     *        defined on deployment {@link EngineFactory} that 'instantiated' the application
     * @param componentLoader
     *        defined on deployment from which runtime application definition, i.e. as modeled
     *        by {@link ApplicationRT}, was loaded
     */
    public ApplicationClassLoader(String id, URL[] urls, DeploymentClassLoader factoryLoader,
            DeploymentClassLoader componentLoader) {
        super(id, urls);
        if (factoryLoader == null) {
            throw new IllegalArgumentException("factoryLoader null");
        } else {
            this.factoryLoader = factoryLoader;
        }
        if (componentLoader == null) {
            throw new IllegalArgumentException("componentLoader null");
        } else {
            this.componentLoader = componentLoader;
        }
    }

    /**
     * Alternative constructor, i.e. in instances where sun.misc.Launcher$AppClassLoader cannot
     * be configured by <code>System</code> with 'provided dependencies.'
     * 
     * @param id
     *        identifier which SHOULD be unique within context of JVM
     * @param urls
     *        the URLs from which to load classes and resources, i.e. artifacts visible only to
     *        this classloader.
     * @param parent
     *        the parent class loader for delegation
     * @param factoryLoader
     *        defined on deployment containing factory that loaded <code>Application</code>
     * @param componentLoader
     *        defined on deployment from which <code>Application</code> definition was
     *        loaded.
     */
    public ApplicationClassLoader(String id, URL[] urls, ClassLoader parent, DeploymentClassLoader factoryLoader,
            DeploymentClassLoader componentLoader) {
        super(id, urls, parent);
        if (factoryLoader == null) {
            throw new IllegalArgumentException("factoryLoader null");
        } else {
            this.factoryLoader = factoryLoader;
        }
        if (componentLoader == null) {
            throw new IllegalArgumentException("componentLoader null");
        } else {
            this.componentLoader = componentLoader;
        }
    }

    /**
     * Gets identifier which is unique within context of JVM. Returns qualified name of runtime
     * engine which implements application serialized as a string.
     * 
     * @return String
     */
    @Override
    public String getID() {
        return super.getID();
    }

    /*
     * (non-Javadoc)
     * @see java.lang.ClassLoader#loadClass(java.lang.String, boolean)
     * 
     * NOTE: we synchronize to avoid race condition where two threads attempt to load same
     * class simultaneously (method on super is synchronized, but we need to lock method on
     * super AND on factory loader AND on component loader) if class not loaded both threads
     * will invoke findClass. it is then a race in terms of which one will load and define
     * class and which one will throw a class linkage error:
     * 
     * "attempted duplicate class definition for name ..."
     */
    @Override
    protected synchronized Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {

        try {

            // delegate to parent classloader. if not found
            // invokes find class on URLClassLoader which will
            // search urls scoped to this classloader

            return super.loadClass(name, resolve);

        } catch (ClassNotFoundException cf) {

            // delegate to factory classloader. note that we
            // invoke loadClassInternal which only checks its
            // repository, i.e. no shared loader delegation

            try {

                return factoryLoader.loadClassInternal(name);

            } catch (ClassNotFoundException cfe) {

                // delegate to component classloader. note that we
                // invoke loadClass method which will query the
                // shared classloader after checking its repo. if
                // not found, it throws a cfe

                return componentLoader.loadClass(name);

            }

        }

    }

    /*
     * (non-Javadoc)
     * @see java.lang.ClassLoader#loadClass(java.lang.String)
     */
    @Override
    public Class<?> loadClass(String name) throws ClassNotFoundException {
        return this.loadClass(name, false);
    }

    /*
     * (non-Javadoc)
     * @see java.lang.ClassLoader#getResource(java.lang.String)
     */
    @Override
    public URL getResource(String name) {
        Enumeration<URL> answer = null;
        try {
            answer = getResources(name);
        } catch (IOException ie) {
            throw new RuntimeException(ie);
        }
        if (answer.hasMoreElements()) {
            return answer.nextElement();
        }
        return null;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.ClassLoader#getResources(java.lang.String)
     * 
     * NOTE: we synchronize to avoid race condition where two threads attempt to load same
     * class simultaneously (method on super is synchronized, but we need to lock method on
     * super AND on factory loader AND on component loader) if class not loaded both threads
     * will invoke findClass. it is then a race in terms of which one will load and define
     * class and which one will throw a class linkage error:
     * 
     * "attempted duplicate class definition for name ..."
     */
    @Override
    public Enumeration<URL> getResources(String name) throws IOException {

        // invoke super first to find resources scoped to parent
        // classloader. if not found super invokes findResources
        // on URLClassLoader which checks our repo

        Enumeration<URL> answer = super.getResources(name);

        // if nothing found, delegate to factory classloader. note
        // that we invoke getResourcesInternal which only checks its
        // repository, i.e. no shared loader delegation

        if (!answer.hasMoreElements()) {
            answer = factoryLoader.getResourcesInternal(name);
        }

        // if nothing found, delegate to factory classloader. note
        // that we invoke getResources method which will query the
        // shared classloader after checking its repo

        if (!answer.hasMoreElements()) {
            answer = componentLoader.getResources(name);
        }

        if (!answer.hasMoreElements()) {
            return answer;
        }

        // fixes a bug within jar handler that returns a jar entry when searching
        // for meta-inf/services/foo and foo is a directory rather than a file. if
        // a dir entry, the returned stream is unusable. detect error by attempting
        // to close stream. as of jdk 1.5.0_08

        Vector<URL> validURLs = new Vector<URL>();
        while (answer.hasMoreElements()) {
            URL next = answer.nextElement();
            if (next.getProtocol().equals("jar")) {
                try {
                    next.openStream().close();
                    validURLs.add(next);
                } catch (Exception e) {
                }
            } else {
                validURLs.add(next);
            }
        }
        return validURLs.elements();

    }
    
    /*
     * (non-Javadoc)
     * @see java.net.URLClassLoader#getURLs()
     */
    @Override
    public URL[] getURLs() {
        URL[] a = factoryLoader.getURLsInternal();
        URL[] b = componentLoader.getURLs();
        URL[] c = super.getURLs();
        URL[] combined = new URL[a.length + b.length + c.length];
        System.arraycopy(a, 0, combined, 0, a.length);
        System.arraycopy(b, 0, combined, a.length, b.length);
        System.arraycopy(c, 0, combined, a.length + b.length, c.length);
        return combined;
    }

}
