/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.io.File;

/**
 * Models a dependency, i.e. an 'artifact' upon which another 'artifact' depends in order to
 * implement its behavior.
 * 
 * @see DependencyManager
 */
public interface Dependency {

    /**
     * Returns ref URI which serves as a unique identifier for this dependency. The value
     * SHOULD be globally unique. The format of ref URI MUST be:
     * <p>
     * organizationID/artifactID/version/extension
     */
    public String getRef();

    /**
     * Returns same value as {@link #getRef()}, but sans version. The format of versionless
     * URI MUST be:
     * <p>
     * organizationID/artifactID/extension
     * <p>
     * This URI MAY be used to detect dependency conflicts, i.e. when more than one version of
     * the same artifact is specified by two or more artifacts.
     * 
     * @return String
     */
    public String getVersionlessRef();

    /**
     * Returns 'short' name which MAY be useful for logging, i.e. when uniqueness is not a
     * requirement and ref URI is unwieldy. Format of short name MUST be:
     * <p>
     * artifactID:extension:version
     * 
     * @return String
     */
    public String getShortName();

    /**
     * Returns an identifier which identifies the organization which authored artifact.
     * 
     * @return String
     */
    public String getOrganizationID();

    /**
     * Returns name which uniquely identifies dependency within context of organization.
     * 
     * @return String
     * @see Deployment#getArtifactID()
     */
    public String getArtifactID();

    /**
     * Returns artifact version, i.e. revision.
     * 
     * @return String
     */
    public String getVersion();

    /**
     * Returns artifact extension, e.g. jar, eoa-factory, eoa-feature, eoa-component, etc ...
     * 
     * @return String
     * 
     */
    public String getExtension();

    /**
     * Returns artifact file.
     * 
     * @return File
     */
    public File getFile();

}
