/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * Models an EOA <code>Deployment</code>. A <code>Deployment</code> contains a single
 * {@link ExtensionFactory} provider definition or one or more {@link RootComponent} and/or
 * {@link RootComponentRT} definitions. The file structure MUST be as follows (note that names
 * are case sensitive):
 * <p>
 * <li>EOA-INF/ (optional)
 * <li>&nbsp;&nbsp;&nbsp; lib/ (optional)
 * <li>META-INF/
 * <li>&nbsp;&nbsp;&nbsp; eoa/deployment.xml (required)
 * <p>
 */
public abstract class Deployment {

    public static final String DESCRIPTOR_NS = "http://bluestemsoftware.org/specification/eoa/1.0/deployment";
    public static final String DESCRIPTOR_NS_PREFIX = "eoa";
    public static final String DESCRIPTOR_PATH = "META-INF/eoa/deployment.xml";
    public static final String DESCRIPTOR_SCHEMA_PATH = "schema/http.bluestemsoftware.org.specification.eoa.1.0.deployment.xsd";

    protected File file;
    protected String organizationID;
    protected String artifactID;
    protected String version;
    protected String extension;
    protected DeploymentContext deploymentContext;
    protected DeploymentDependencies deploymentDependencies;
    protected Date deployed;
    
    /**
     * Constructor.
     * 
     * @param organizationID
     * @param artifactID
     * @param version
     * @param extension
     * @param file
     *        which MUST be a directory, i.e. if file is an archive, the file MUST be unzipped
     *        to a temporary location unless deployment loaded from classpath.
     * @param deploymentDependencies
     */
    public Deployment(String organizationID, String artifactID, String version, String extension, File file,
            DeploymentDependencies deploymentDependencies) {
        if (organizationID == null) {
            throw new IllegalArgumentException("organizationID null");
        }
        if (artifactID == null) {
            throw new IllegalArgumentException("artifactID null");
        }
        if (version == null) {
            throw new IllegalArgumentException("version null");
        }
        if (extension == null) {
            throw new IllegalArgumentException("extension null");
        }
        if (file == null) {
            throw new IllegalArgumentException("file null");
        }
        if (artifactID == null) {
            throw new IllegalArgumentException("artifactID null");
        }
        if (deploymentDependencies == null) {
            throw new IllegalArgumentException("deploymentDependencies null");
        }
        this.organizationID = organizationID;
        this.artifactID = artifactID;
        this.version = version;
        this.extension = extension;
        this.file = file;
        this.deploymentDependencies = deploymentDependencies;
    }

    /**
     * Gets id of <code>Deployment</code> 'artifact.'
     * 
     * @return artifactID
     */
    public String getArtifactID() {
        return artifactID;
    }

    /**
     * Gets deployment 'extension', e.g. eoa-factory, eoa-feature or eoa-component.
     * 
     * @return extension
     */
    public String getExtension() {
        return extension == null ? "" : extension;
    }

    /**
     * Gets value which identifies 'organization' that authored <code>Deployment</code>.
     * 
     * @return organizationID
     */
    public String getOrganizationID() {
        return organizationID == null ? "" : organizationID;
    }

    /**
     * Gets <code>Deployment</code> version.
     * 
     * @return version
     */
    public String getVersion() {
        return version == null ? "" : version;
    }

    /**
     * Returns ref URI which serves as a unique identifier for this deployment. The value MUST
     * be unique within contet of <code>System</code>.
     */
    public abstract String getRef();

    /**
     * Returns <code>Deployment</code> file which is a directory.
     * 
     * @return File
     */
    public File getFile() {
        return file;
    }

    /**
     * Returns date deployment was deployed.
     * 
     * @return deployment date, or <code>null</code> if deployment not yet deployed.
     */
    public Date getDateDeployed() {
        return deployed;
    }

    /**
     * Returns dependencies defined by this <code>Deployment</code>.
     * 
     * @return DeploymentDependencies
     */
    public DeploymentDependencies getDependencies() {
        return deploymentDependencies;
    }

   protected List<Element> getChildElements(Element parent, QName childName) {
        List<Element> answer = new ArrayList<Element>();
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                if (element.getNamespaceURI() != null) {
                    continue;
                }
            } else {
                if (element.getNamespaceURI() == null) {
                    continue;
                }
                if (!childName.getNamespaceURI().equals("*")
                        && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                    continue;
                }
            }
            if (!childName.getLocalPart().equals("*") && !element.getLocalName().equals(childName.getLocalPart())) {
                continue;
            }
            answer.add(element);
        }
        return answer;
    }

    protected Element getChildElement(Element parent, QName childName) {
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                if (element.getNamespaceURI() != null) {
                    continue;
                }
            } else {
                if (element.getNamespaceURI() == null) {
                    continue;
                }
                if (!childName.getNamespaceURI().equals("*")
                        && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                    continue;
                }
            }
            if (!childName.getLocalPart().equals("*") && !element.getLocalName().equals(childName.getLocalPart())) {
                continue;
            }
            return element;
        }
        return null;
    }

    protected String getText(Element element) {
        if (element == null || !element.hasChildNodes()) {
            return null;
        }
        try {
            Text textNode = (Text)element.getFirstChild();
            return textNode.getData();
        } catch (ClassCastException ce) {
        }
        return null;
    }

    protected class DescriptorParser implements ErrorHandler {

        private ArrayList<String> errorMessages = new ArrayList<String>();
        private DocumentBuilderFactory factory;

        public DescriptorParser() throws Exception {
            factory = DocumentBuilderFactory.newInstance();
            factory.setNamespaceAware(true);
            factory.setSchema(DeploymentManager.getDescriptorSchema());
        }

        public Document parse(InputSource inputSource) throws SAXException, IOException {
            errorMessages.clear();
            DocumentBuilder builder;
            try {
                builder = factory.newDocumentBuilder();
            } catch (ParserConfigurationException pe) {
                throw new SAXException(pe);
            }
            builder.setErrorHandler(this);
            builder.setEntityResolver(deploymentContext);
            Document document = builder.parse(inputSource);
            if (errorMessages.size() > 0) {
                Iterator<String> messageIterator = errorMessages.iterator();
                StringBuilder errorMessage = new StringBuilder();
                while (messageIterator.hasNext()) {
                    errorMessage.append(messageIterator.next() + "  ");
                }
                throw new SAXException(errorMessage.toString());
            }
            return document;
        }

        public void error(SAXParseException spe) throws SAXException {
            errorMessages.add(spe.getMessage());
        }

        public void warning(SAXParseException spe) throws SAXException {
            errorMessages.add(spe.getMessage());
        }

        public void fatalError(SAXParseException spe) throws SAXException {
            errorMessages.add(spe.getMessage());
        }

    }

}