/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.io.IOException;
import java.io.StringReader;
import java.net.URLStreamHandler;
import java.util.WeakHashMap;

import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Models context within which a deployment is deployed. Deployment is either an
 * <code>ExtensionFactoryDeployment</code> or a <code>ComponentDeployment</code>.
 * <p>
 * Note that an instance of this class MAY be used as an entity caching SAX2
 * <code>EntityResolver</code>. Instances of <code>DeploymentContext</code> are
 * threadsafe.
 */
public abstract class DeploymentContext implements EntityResolver {

    protected DeploymentClassLoader classLoader;
    protected WeakHashMap<String, String> entityCache = new WeakHashMap<String, String>();

    public DeploymentContext(DeploymentClassLoader classLoader) {
        if (classLoader == null) {
            throw new IllegalArgumentException("classLoader null");
        }
        this.classLoader = classLoader;
    }

    /**
     * Returns <code>ClassLoader</code> which utilizes class repository scoped to
     * <code>Deployment</code> which MUST be separate and apart from repository utilized by
     * other <code>Deployment</code> objects.
     * 
     * @return ClassLoader Deployment classloader
     */
    public DeploymentClassLoader getClassLoader() {
        return classLoader;
    }

    /**
     * Returns resource at indicated location. Uses classloader scoped to this deployment if
     * retrieving a resource located on classpath. A <code>Container</code> MAY provide
     * instances of {@link URLStreamHandler} in place of default JDK handlers.
     * 
     * @param location
     *        URL in string format. he supported schemes are as follows. Note that for
     *        portability reasons, file scheme SHOULD NOT be used unless file is
     *        located within deployment.
     *        <p>
     *        http:
     *        <p>
     *        https:
     *        <p>
     *        ftp:
     *        <p>
     *        classpath:
     *        <p>
     *        file:
     * 
     * @return Resource which is never <code>null</code>, even when resource is not
     *         dereferenceable. To test for existence of physical resource, check for a
     *         <code>null</code> stream.
     * 
     * @throws DeploymentException
     *         if error occurs retrieving resource.
     */
    public Resource getResource(String location) throws DeploymentException {
        return new Resource(classLoader, location);
    }

    public InputSource resolveEntity(String publicID, String systemID) throws SAXException, IOException {

        String entityString = null;

        if (systemID != null && systemID.length() > 0) {

            String key = null;
            if (entityCache != null) {
                key = publicID + systemID;
                synchronized (entityCache) {
                    entityString = (String)entityCache.get(key);
                }
            }
            if (entityString == null) {
                try {
                    Resource resource = getResource(systemID);
                    entityString = resource.getText("UTF-8", true);
                } catch (DeploymentException se) {
                    throw new IOException("Error resolving systemID '" + systemID + "'. " + se);
                }
                synchronized (entityCache) {
                    entityCache.put(key, entityString);
                }
            }

        } else {
            throw new SAXException("systemID null or an empty string.");
        }

        InputSource inputSource = new InputSource(new StringReader(entityString));
        inputSource.setPublicId(publicID);
        inputSource.setSystemId(systemID);
        return inputSource;

    }

}
