/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.io.File;
import java.io.FileFilter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.xml.transform.sax.SAXSource;

import org.bluestemsoftware.specification.eoa.component.ClasspathDeployment;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactoryContext;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactoryDeployment;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public abstract class DeploymentManager {

    protected static javax.xml.validation.Schema descriptorSchema;
    protected SharedClassLoader sharedClassLoader;

    /**
     * Creates a <code>DeploymentContext</code> object.
     * 
     * @param deployment
     * @return
     * @throws DeploymentException
     */
    public DeploymentContext createDeploymentContext(Deployment deployment) throws DeploymentException {

        DeploymentDependencies dd = deployment.getDependencies();

        List<URL> constituents = new ArrayList<URL>();

        if (deployment instanceof ClasspathDeployment == false) {

            // add dependencies scoped to deployment to classpath. no need to
            // worry about version management. dependency manager handled this

            for (ScopedDependency sd : dd.getScopedDependencies()) {
                String decodedPath = sd.getFile().toURI().getSchemeSpecificPart();
                try {
                    constituents.add(new File(decodedPath).toURI().toURL());
                } catch (MalformedURLException me) {
                    throw new DeploymentException(me.toString());
                }
            }

            // add all jars defined within optional lib subdirectory to
            // classpath

            File libDir = new File(deployment.getFile(), "EOA-INF/lib/");
            if (libDir.exists() && libDir.isDirectory()) {
                File[] contents = libDir.listFiles(new FileFilter() {
                    public boolean accept(File pathname) {
                        return pathname.toString().endsWith("jar");
                    }
                });
                if (contents.length > 0) {
                    for (int i = 0; i < contents.length; i++) {
                        String decodedPath = contents[i].toURI().getSchemeSpecificPart();
                        try {
                            constituents.add(new File(decodedPath).toURI().toURL());
                        } catch (MalformedURLException me) {
                            throw new DeploymentException(me.toString());
                        }
                    }
                }
            }

            // add optional 'classes.jar' to the deployment classpath

            File classesLib = new File(deployment.getFile(), "EOA-INF/classes.jar");
            if (classesLib.exists()) {
                String decodedPath = classesLib.toURI().getSchemeSpecificPart();
                try {
                    constituents.add(new File(decodedPath).toURI().toURL());
                } catch (MalformedURLException me) {
                    throw new DeploymentException(me.toString());
                }
            }

        }

        // use shared classloader's parent as parent of deployment classloader which
        // should be java system classloader. note that if parent is a shared
        // classloader instance, this is special case set-up within our test
        // harness, i.e. tests shared classloader extends shared classloader and 
        // actual shared classloader is parent

        URL[] scopedArtifacts = new URL[constituents.size()];
        scopedArtifacts = (URL[])constituents.toArray(scopedArtifacts);
        ClassLoader parent = null;
        if (sharedClassLoader.getParent() instanceof SharedClassLoader) {
            parent = sharedClassLoader.getParent().getParent();
        } else {
            parent = sharedClassLoader.getParent();
        }

        DeploymentClassLoader dcl = new DeploymentClassLoader(deployment.getRef(), scopedArtifacts, parent,
                sharedClassLoader);

        if (deployment instanceof ComponentDeployment) {
            return new ComponentContext((ComponentDeployment)deployment, dcl);
        } else {
            return new ExtensionFactoryContext((ExtensionFactoryDeployment)deployment, dcl);
        }

    }

    public static javax.xml.validation.Schema getDescriptorSchema() throws SAXException {
        if (descriptorSchema == null) {
            javax.xml.validation.SchemaFactory schemaFactory = null;
            schemaFactory = javax.xml.validation.SchemaFactory.newInstance("http://www.w3.org/2001/XMLSchema");
            try {
                InputSource inputSource = new InputSource();
                ClassLoader cl = Thread.currentThread().getContextClassLoader();
                inputSource.setByteStream(cl.getResourceAsStream(Deployment.DESCRIPTOR_SCHEMA_PATH));
                inputSource.setSystemId(Deployment.DESCRIPTOR_SCHEMA_PATH);
                descriptorSchema = schemaFactory.newSchema(new SAXSource(inputSource));
            } catch (Exception ex) {
                if (ex instanceof SAXException) {
                    throw (SAXException)ex;
                }
                throw new SAXException("Error creating descriptor grammar. " + ex);
            }
        }
        return descriptorSchema;
    }

}
