/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.util.List;

/**
 * Identifies a 'factory' dependency, i.e. the referencing artifact (system) depends upon the
 * ability to retrieve and utilize referenced <code>ExtensionFactory</code> in order to
 * implement its behavior. Note that <code>FactoryDependency</code> artifacts MUST NOT be
 * added as classpath constituents.
 * <p>
 * The rules used to distinguish a <code>FactoryDependency</code> from other dependencies are
 * an extension to the core specification.
 */
public interface FactoryDependency extends Dependency {

    /**
     * Identifies 'provided' dependencies upon which this <code>ExtensionFactory</code>
     * depends. Transitive dependencies, i.e. dependencies upon which a provided dependency MAY
     * depend, MUST NOT be included. Only dependencies declared explicitly within referencing
     * project are returned.
     * 
     * @return List of provided dependencies declared directly within
     *         <code>ExtensionFactory</code> project, or an empty list if none defined.
     */
    public List<ProvidedDependency> getProvidedDependencies();

    /**
     * Identifies 'shared' dependencies upon which this <code>ExtensionFactory</code>
     * depends. Transitive dependencies, i.e. dependencies upon which a shared dependency MAY
     * depend, MUST NOT be included. Only dependencies declared explicitly within referencing
     * project are returned.
     * 
     * @return List of shared dependencies declared directly within
     *         <code>ExtensionFactory</code> project, or an empty list if none defined.
     */
    public List<SharedDependency> getSharedDependencies();

    /**
     * Identifies 'scoped' dependencies upon which this <code>ExtensionFactory</code>
     * depends. Transitive dependencies, i.e. dependencies upon which a scoped dependency MAY
     * depend MUST be processed according to type specific rules and added to list as
     * indicated.
     * 
     * @return List of scoped dependencies or an empty list if none defined
     */
    public List<ScopedDependency> getScopedDependencies();

    /**
     * Identifies <code>Feature</code> definitions upon which this
     * <code>ExtensionFactory</code> depends, i.e. this <code>ExtensionFactory</code>
     * depends upon referenced <code>Feature</code> to implement its behavior.
     * <p>
     * Only dependencies declared explicitly within referencing project are returned.
     * Transitive dependencies, if any, are modeled within each returned dependency.
     * 
     * @return List<FeatureDependency> or an empty list if no dependencies defined
     */
    public List<FeatureDependency> getFeatureDependencies();

}
