/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;

/**
 * Uniquely identifies a component model 'fragment', the syntax of which is compliant with the
 * xpointer framework. When serialized as a string, expression is formatted as follows:
 * <p>
 * rootComponentNamespace#scheme.componentType(pointerPart)
 * <p>
 * e.g.
 * <p>
 * http://mycompany.com/eoa/component/1.0#eoa.engine(myengine)
 */
public class FragmentIdentifier {

    public static enum Scheme {
        EOA
    }

    private String rootComponentNamespace;
    private Scheme scheme;
    private String componentType;
    private String pointerPart;
    private String expression;
    private String schemeSpecificPart;

    private static Pattern pattern = Pattern.compile("(.+)[#](eoa)[.](\\w+)[(](.*)[)]");

    /**
     * Constructs a <code>FragmentIdentifier</code>.
     * 
     * @param rootComponentNamespace
     *        target namespace of root component
     * @param scheme
     *        enumerated value
     * @param componentType
     *        a label which identifies type of component
     * @param pointerPart
     *        scheme specific part sans beginning and ending paren or <code>null</code> if
     *        pointer part empty.
     */
    public FragmentIdentifier(String rootComponentNamespace, Scheme scheme, String componentType,
            String pointerPart) {
        super();
        if (rootComponentNamespace == null) {
            throw new IllegalArgumentException("rootComponentNamespace null");
        } else {
            this.rootComponentNamespace = rootComponentNamespace;
        }
        if (scheme == null) {
            throw new IllegalArgumentException("scheme null");
        } else {
            this.scheme = scheme;
        }
        if (componentType == null) {
            throw new IllegalArgumentException("componentType null");
        } else {
            this.componentType = componentType;
        }
        this.pointerPart = pointerPart;
    }

    public String getRootComponentNamespace() {
        return rootComponentNamespace;
    }

    /**
     * Gets point part.
     * @return String or <code>null</code> if point part empty.
     */
    public String getPointerPart() {
        return pointerPart;
    }

    public Scheme getScheme() {
        return scheme;
    }

    public String getComponentType() {
        return componentType;
    }

    /**
     * Gets scheme specific part, i.e. character sequence following scheme, e.g. the scheme
     * specific part for the fragment identifier 'http://foo#eoa.binding(mybinding)' would be
     * 'binding(mybinding)'.
     * 
     * @return
     */
    public String getSchemeSpecificPart() {
        if (schemeSpecificPart == null) {
            StringBuilder sb = new StringBuilder();
            sb.append(componentType);
            sb.append("(");
            sb.append(pointerPart);
            sb.append(")");
            schemeSpecificPart = sb.toString();
        }
        return schemeSpecificPart;
    }

    public static FragmentIdentifier valueOf(String expression) {
        if (expression == null) {
            throw new IllegalArgumentException("expression null");
        }
        Matcher matcher = pattern.matcher(expression);
        if (matcher.find()) {
            String rootNamespace = matcher.group(1);
            Scheme scheme = Scheme.valueOf(matcher.group(2).toUpperCase());
            String componentType = matcher.group(3);
            String pointerPart = matcher.group(4);
            return new FragmentIdentifier(rootNamespace, scheme, componentType, pointerPart);
        } else {
            throw new IllegalArgumentException("FragmentIdentifier expression formatted incorrectly. "
                    + expression);
        }
    }

    public static FragmentIdentifier valueOf(ComponentType type, QName name) {
        if (type == null) {
            throw new IllegalArgumentException("type null");
        }
        if (name == null) {
            throw new IllegalArgumentException("name null");
        }
        return new FragmentIdentifier(name.getNamespaceURI(), Scheme.EOA, type.toString().toLowerCase(), name
                .getLocalPart());
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        if (expression == null) {
            StringBuilder sb = new StringBuilder();
            sb.append(rootComponentNamespace);
            sb.append("#");
            sb.append(scheme.toString().toLowerCase());
            sb.append(".");
            sb.append(getSchemeSpecificPart());
            expression = sb.toString();
        }
        return expression;
    }

    @Override
    public final int hashCode() {
        return hashCode(toString());
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof FragmentIdentifier)) {
            return false;
        }

        if ((toString() == ((FragmentIdentifier)obj).toString() || (toString() != null && toString().equals(
                ((FragmentIdentifier)obj).toString())))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
