/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component;

import java.util.HashSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment.CurrentReference;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT;
import org.w3c.dom.Node;

/**
 * Models a 'top-level' component defined at the 'root' of the component model hierarchy. Root
 * components are not owned by other components, but are instead referenced, i.e. as a Service
 * references an Interface.
 */

public abstract class RootComponent implements Component {

    /**
     * Defines namespace for components which are 'built-in' to this specification. Note that
     * this namespace is reserved and MUST NOT be used to name components other than those
     * designated by this specification.
     */
    public static final String EOA_COMPONENT_NS = "http://bluestemsoftware.org/specification/eoa/component";

    protected ComponentName name;
    protected ComponentContext componentContext;
    protected FragmentIdentifier fragmentIdentifier;
    protected RootComponentRT runtimeProvider;
    protected Set<ComponentReference<?>> references = new HashSet<ComponentReference<?>>();

    private Boolean isDeployed = Boolean.FALSE;

    /**
     * Enumerates root component types
     */
    public static enum ComponentType {
        APPLICATION, ENGINE, SERVICE, BINDING, INTERFACE, MESSAGE, SCHEMA, POLICY
    }

    /**
     * Identifies a specific type of <code>RootComponent</code>, i.e. a component which
     * supports a runtime manifestation of self. The runtime component 'provides' runtime
     * behavior.
     */
    public interface RuntimeProvidable {

        /**
         * Returns runtime manifestation of self. Note that definition for runtime component
         * MUST be defined within same deployment as definition for corresponding
         * <code>RootComponent</code>.
         * 
         * @return RootComponentRT
         * @throws IllegalStateException
         *         if provider definition not yet deployed.
         */
        public RootComponentRT getRuntimeProvider();

    }

    /**
     * Gets qualified name of component. Name MUST be unique within the context of
     * <code>System</code> for indicated <code>ComponentType</code>.
     * 
     * @return ComponentName Qualified name of component
     */
    public ComponentName getName() {
        return name;
    }

    /**
     * Gets component type.
     * 
     * @return ComponentType
     */
    public abstract ComponentType getComponentType();

    /**
     * Indicates if <code>RootComponent</code> definition is 'built-in' to this
     * specification.
     * 
     * @return true if <code>RootComponent</code> is 'built-in', false if not.
     */
    public boolean isBuiltIn() {
        if (name.getNamespaceURI().equals(EOA_COMPONENT_NS)) {
            return true;
        }
        if (name.getNamespaceURI().equals("http://www.w3.org/2001/XMLSchema")) {
            return true;
        }
        return false;
    }

    /**
     * Gets <code>ComponentContext</code> which models context within which a
     * <code>RootComponent</code> is deployed.
     * 
     * @return ComponentContext
     */
    public ComponentContext getComponentContext() {
        return componentContext;
    }

    /*
     * Invoked by <code>ComponentDeployment</code>. Invokes {@link #establishReferences()}
     * and {@link #validate()}.
     * 
     * @param componentContext @throws DeploymentException
     */
    synchronized final void deploy(ComponentContext componentContext) throws DeploymentException {

        if (isDeployed == null) {
            throw new DeploymentException("RootComponent "
                    + getFragmentIdentifier()
                    + " is circularly referenced by "
                    + CurrentReference.getCurrentReference());
        } else {
            isDeployed = null;
        }

        this.componentContext = componentContext;
        establishReferences();
        validate();

        isDeployed = Boolean.TRUE;

    }

    /**
     * Indicates whether <code>RootComponent</code> definition has been deployed, i.e.
     * references have been established and component definition has been validated. A
     * definition is considered to be undefined until it has been successfully deployed.
     * 
     * @return true if deployed, false if not.
     */
    public synchronized boolean isDeployed() {
        return isDeployed == null ? false : isDeployed.booleanValue();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        return fragmentIdentifier;
    }

    /*
     * (non-Javadoc)
     * 
     * Forces subclass to establish references made by this component or by components nested
     * within this component.
     */
    protected abstract void establishReferences() throws DeploymentException;

    /*
     * (non-Javadoc)
     * 
     * Invoked after component has been deployed, i.e. after all component references have been
     * established. Validates that component structure is valid according to specification and
     * that referenced components are compatible. <p> Invok
     */
    protected abstract void validate() throws DeploymentException;

    /*
     * (non-Javadoc)
     * 
     * Used by ComponentDeployment to set runtime component. Note that we define method here as
     * opposed to a public method on RuntimeProvidable such that we can restrict access to
     * ComponentManager. Consequently, if RootComponent is not an instanceof RuntimeProvidable,
     * this method is not used.
     */
    void setRuntimeProvider(RootComponentRT runtimeProvider) throws DeploymentException {
        if (this.runtimeProvider != null) {
            throw new DeploymentException("RuntimeProvider already defined for RootComponent "
                    + getFragmentIdentifier());
        }
        this.runtimeProvider = runtimeProvider;
    }

    /*
     * this method is overridden by interface component and is invoked by ComponentDeployment
     * after provider definitions have been deployed, i.e. we need runtime schema to generate
     * the models. This is a unique case because interface component is static, i.e. it has no
     * runtime provider (which typically performs validation of self including validation
     * against static component definition, i.e. RuntimeProvidable.)
     */
    protected void modelOperationStyles() throws DeploymentException {
    }

    /**
     * Qualified name of <code>RootComponent</code>.
     * 
     * @see #valueOf(Node, String)
     */
    public static class ComponentName extends QName {

        private static final long serialVersionUID = 1L;

        private static final Pattern prefixedNameRegExp = Pattern.compile("((\\S+):){0,}(\\w+)");

        public ComponentName(String namespaceURI, String localPart) {
            super(namespaceURI, localPart);
        }

        private ComponentName(String namespaceURI, String localPart, String prefix) {
            super(namespaceURI, localPart, prefix);
        }

        /**
         * Utility method which parses a prefixed name within context of an element into a
         *  <code>ComponentName</code>.
         * 
         * @param parent
         *        node containing in scope namespace declarations.
         * @param prefixedName
         *        Note that the local part and prefix are not validated as an <a
         *        href="http://www.w3.org/TR/REC-xml-names/#NT-NCName">NCName</a> as specified
         *        in <a href="http://www.w3.org/TR/REC-xml-names/">Namespaces in XML</a>.
         *        </p>
         * @return ComponentName
         */
        public static ComponentName valueOf(Node parent, String prefixedName) {
            if (parent == null) {
                throw new IllegalArgumentException("parent null");
            }
            if (prefixedName == null || prefixedName.equals("")) {
                throw new IllegalArgumentException("prefixedName null or empty string");
            }
            String prefix = getPrefix(prefixedName);
            String localName = getLocalPart(prefixedName);
            String namespaceURI = null;
            if (prefix != null && prefix.equals("") == false) {
                namespaceURI = parent.lookupNamespaceURI(prefix);
                if (namespaceURI == null || namespaceURI.equals("")) {
                    throw new UnresolvedPrefixException("Error parsing prefixed name '"
                            + prefixedName
                            + "' into a QName. Namespace prefix '"
                            + prefix
                            + "' failed to resolve to a namespace uri.");
                }
            }
            if (prefix == null) {
                return new ComponentName(namespaceURI, localName);
            } else {
                return new ComponentName(namespaceURI, localName, prefix);
            }
        }

        private static String getPrefix(String qualifiedName) {
            if (qualifiedName != null) {
                Matcher matcher = prefixedNameRegExp.matcher(qualifiedName);
                if (matcher.find()) {
                    return matcher.group(2);
                } else {
                    return null;
                }
            } else {
                return null;
            }
        }

        private static String getLocalPart(String qualifiedName) {
            if (qualifiedName != null) {
                Matcher matcher = prefixedNameRegExp.matcher(qualifiedName);
                if (matcher.find()) {
                    return matcher.group(3);
                } else {
                    return null;
                }
            } else {
                return null;
            }
        }

        static class UnresolvedPrefixException extends RuntimeException {
            private static final long serialVersionUID = 1L;

            public UnresolvedPrefixException(String message) {
                super(message);
            }
        }

        static class QNameFormatException extends RuntimeException {
            private static final long serialVersionUID = 1L;

            public QNameFormatException(String message) {
                super(message);
            }
        }

    }

}