/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.application.rt;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.Application;
import org.bluestemsoftware.specification.eoa.component.application.Role;
import org.bluestemsoftware.specification.eoa.system.SystemContext;

/**
 * An <code>ApplicationReference</code> implementation which MAY be extended by type specific
 * descendants as required.
 */
public abstract class AbstractApplicationReference implements ApplicationReference {

    protected FragmentIdentifier fragmentIdentifier;
    protected boolean isMyApplication;
    protected ComponentName applicationName;
    protected ApplicationRT parent;
    protected Application application;
    protected Map<String, RoleReference> roleReferences;

    public AbstractApplicationReference(ComponentName applicationName, Map<String, RoleReference> roleReferences,
            boolean isMyApplication) {
        if (applicationName == null) {
            throw new IllegalArgumentException("applicationName null");
        } else {
            this.applicationName = applicationName;
        }
        if (roleReferences == null) {
            throw new IllegalArgumentException("roleReferences null");
        } else {
            this.roleReferences = roleReferences;
            for (RoleReference roleReference : roleReferences.values()) {
                ((AbstractRoleReference)roleReference).setParent(this);
            }
        }
        this.isMyApplication = isMyApplication;
    }

    /**
     * Gets a <code>RoleReference</code> component.
     * 
     * @param roleName
     * @return
     */
    public RoleReference getRoleReference(String roleName) {
        return roleReferences.get(roleName);
    }

    /**
     * Gets all <code>RoleReference</code> components.
     * 
     * @return
     */
    public Collection<RoleReference> getRoleReferences() {
        return roleReferences.values();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getParent()
     */
    public ApplicationRT getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public ApplicationRT getRootComponent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public Application getReferencedComponent() {
        return application;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return applicationName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<Application> getReferencedComponentClass() {
        return Application.class;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.APPLICATION;
    }

    protected void setParent(ApplicationRT parent) {
        this.parent = parent;
    }

    protected void establishReference() throws DeploymentException {
        if (application != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        if (isMyApplication) {
            application = (Application)getParent().getRuntimeProvidable();
        } else {
            try {
                application = getRootComponent().getComponentContext().findReferencedComponent(this);
            } catch (DeploymentException de) {
                throw new DeploymentException("Application reference "
                        + getFragmentIdentifier()
                        + " defined within deployment "
                        + getRootComponent().getComponentContext().getDeployment().getRef()
                        + " is invalid. "
                        , de);
            }
            if (application == null) {
                throw new DeploymentException("ApplicationReference "
                        + getFragmentIdentifier()
                        + " is invalid. Referenced Application "
                        + applicationName
                        + " is undefined.");
            }
        }
        generateImpliedStructure();
        for (RoleReference rr : roleReferences.values()) {
            ((AbstractRoleReference)rr).establishReference();
        }
    }

    /**
     * Used to generate implied structure, i.e. <code>RoleReference</code> components not
     * explicitly defined within provider definition.
     * 
     * @param roleName
     * @return
     */
    protected abstract AbstractRoleReference createRoleReference(String roleName) throws DeploymentException;

    protected void generateImpliedStructure() throws DeploymentException {
        Application app = SystemContext.getContext().getSystem().getApplication(application.getName());
        for (Role role : app.getRoles()) {
            if (roleReferences.get(role.getName()) == null) {
                AbstractRoleReference rr = createRoleReference(role.getName());
                rr.setParent(this);
                roleReferences.put(role.getName(), rr);
            }
        }
        roleReferences = Collections.unmodifiableMap(roleReferences);
    }

    protected void validate() throws DeploymentException {
        // TODO: verify that our structure mirrors structure defined on referenced
        // application
    }

    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(rootName.getLocalPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(applicationName.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + applicationName.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "applicationReference", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("applicationName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(applicationName);
        sb.append(System.getProperty("line.separator"));
        sb.append("isMyApplication: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(isMyApplication);
        sb.append(System.getProperty("line.separator"));
        sb.append("roleReferences: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(roleReferences);
        return sb.toString();
    }

}
