/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.binding.rt;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.binding.BindingFault;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;

/**
 * Encapsulates information related to an error which ocurred while processing an
 * <code>Action</code> within 'binding' layer.
 * <p>
 * Note that this is not a runtime manifestation of {@link BindingFault} which binds a declared
 * abstract fault. Instances of <code>BindingFaultRT</code> are undeclared. The 'RT'
 * qualifier is instead intended to distinguish this class from an instance of
 * {@link BindingFault}.
 */
public abstract class BindingFaultRT extends RuntimeException {

    private static final long serialVersionUID = 1L;

    /**
     * A protocol agnostic fault name which indicates error is due to a problem with message
     * sent by sender. This name MAY be used by <code>Binding</code> to name fault when a
     * more specific indication of error is not available, e.g. a SOAP fault code.
     */
    public static final QName SENDER_ERROR = new QName("http://bluestemsoftware.org/specification/eoa",
            "SenderError");

    /**
     * A protocol agnostic fault name which indicates error is due to a problem encountered
     * while receiving message, i.e. as opposed to the message itself. This name MAY be used by
     * <code>Binding</code> to name fault when a more specific indication of error is not
     * available.
     */
    public static final QName RECEIVER_ERROR = new QName("http://bluestemsoftware.org/specification/eoa",
            "ReceiverError");

    protected QName faultName;

    /**
     * Constructor
     * 
     * @param faultReason
     */
    public BindingFaultRT(String faultReason) {
        super(faultReason);
        if (faultReason == null || faultReason.equals("")) {
            throw new IllegalArgumentException("faultReason null");
        }
    }

    /**
     * Constructor
     * 
     * @param faultReason
     * @param th
     *        which caused fault.
     */
    public BindingFaultRT(String faultReason, Throwable th) {
        super(faultReason, th);
        if (faultReason == null || faultReason.equals("")) {
            throw new IllegalArgumentException("faultReason null");
        }
    }

    /**
     * Constructor which sets faultName and faultReason.
     * 
     * @param faultName
     * @param faultReason
     */
    public BindingFaultRT(QName faultName, String faultReason) {
        super(faultReason);
        if (faultName == null) {
            throw new IllegalArgumentException("faultName");
        }
        if (faultReason == null || faultReason.equals("")) {
            throw new IllegalArgumentException("faultReason");
        }
        this.faultName = faultName;
    }

    /**
     * Gets fault name.
     * 
     * @return Fault name or <code>null</code> if undefined.
     */
    public QName getFaultName() {
        return faultName;
    }

    /**
     * Description of error that caused fault.
     * 
     * @return faultReason, which is never <code>null</code>
     */
    public String getFaultReason() {
        return getMessage();
    }

    /**
     * Returns action associated with fault. Note that because binding faults are undeclared,
     * action SHOULD be a static value specific to a concrete instance of this class.
     * 
     * @return String
     */
    public abstract String getAction();

    /**
     * Appends fault detail, if any, to the appropriate parent <code>Element</code> within a
     * protocol specific, e.g. SOAP, version specific, e.g. SOAP 1.1, SOAP 1.2, fault envelope.
     * Note that is responsibility of caller to insure that the parent <code>Element</code>,
     * e.g. SOAP header <code>Element</code> or SOAP fault detail <code>Element</code>,
     * exists prior to invoking this method.
     * <p>
     * Note that detail attached to parent element, if any, MUST be manifested as one or more
     * child <code>Element</code> information items. Parent <code>Element</code> to which
     * detail is appendend MUST NOT be decorated with additional <code>Attribute</code>
     * information items intended to convey detail related to fault.
     * 
     * @param faultEnvelope
     *        Root element of a protocol specific fault envelope whose owner
     *        <code>Document</code> MUST be used to create and/or import child
     *        <code>Element</code> information items.
     */
    public abstract void appendFaultDetail(Content faultEnvelope);

}
