/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.binding.rt;

import java.util.Set;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment.CurrentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.engine.rt.BindingModule;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.TransportModule;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.BusinessFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.FaultContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyExpression;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT.SingletonRT;
import org.bluestemsoftware.specification.eoa.ext.ExecutableException;
import org.bluestemsoftware.specification.eoa.ext.ExecutableExtension;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingException;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.rt.TransportProtocolRT;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.server.BindingFeature;

/**
 * Models a runtime <code>Binding</code> component which binds a type specific messaging
 * protocol to a type specific transport at runtime.
 * <p>
 * Type specific methods, i.e. methods which handle inbound messages delivered by an instance
 * of {@link TransportModule}, are accessed by casting <code>BindingRT</code> to an instance
 * of {@link TransportProtocolRT}, i.e. which defines runtime transport behavior.
 * <p>
 * Note that concrete instances of <code>BindingRT</code> MUST implement exactly one runtime
 * transport protocol, i.e. MUST implement exactly one instance of {@link TransportProtocolRT}.
 */
public abstract class BindingRT extends ExecutableExtension implements SingletonRT<Binding> {

    protected FragmentIdentifier fragmentIdentifier;
    protected ComponentName componentName;
    protected String protocolType;
    protected ComponentContext componentContext;

    private Boolean isDeployed = Boolean.FALSE;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ExecutableExtension.Provider {

        public void spi_init() throws BindingException;

        public void spi_destroy();

        public void spi_suspend() throws ExecutableException;

        public void spi_resume() throws ExecutableException;

        public boolean spi_isSuspended();

        public void spi_validateEndpointReference(EndpointReference endpointReference) throws DeploymentException;

        public Set<String> spi_getRequiredBindingFeatures();

    }

    public BindingRT(BindingFactory factory, Provider provider, ComponentName componentName, String protocolType) {
        super(factory, provider);
        if (componentName == null) {
            throw new IllegalArgumentException("componentName null");
        } else {
            this.componentName = componentName;
        }
        fragmentIdentifier = new FragmentIdentifier(componentName.getNamespaceURI(), Scheme.EOA, "bindingRT",
                componentName.getLocalPart());
        if (protocolType == null) {
            throw new IllegalArgumentException("protocolType null");
        } else {
            this.protocolType = protocolType;
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getRuntimeProvidable()
     */
    public final Binding getRuntimeProvidable() {
        return SystemContext.getContext().getSystem().getBinding(componentName);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ManageableExtension#init()
     */
    public final void init() throws BindingException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_init();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ManageableExtension#destroy()
     */
    @Override
    public final void destroy() {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_destroy();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExecutableExtension#suspend()
     */
    public void suspend() throws ExecutableException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_suspend();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExecutableExtension#resume()
     */
    public void resume() throws ExecutableException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_resume();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExecutableExtension#isSuspended()
     */
    public final boolean isSuspended() {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_isSuspended();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Identifies features which MUST be enaged when <code>Binding</code> processes an
     * action. An exception is thrown during validation of <code>Engine</code> if each listed
     * feature is not enabled on every <code>EndpointActionReference</code> nested within
     * <code>Engine</code> whose referenced <code>Endpoint</code> employs this
     * <code>Binding</code>.
     * <p>
     * Note that if action is the undeclared action {@link SystemFault#ACTION}, required
     * feature is implemented via 'default' <code>BindingModule</code> instance retrieved via
     * {@link BindingFeature#createDefaultBindingModule(boolean)}.
     * 
     * @return a <code>Set</code> of type ref URI's each of which identifies a specific
     *         feature type or an empty <code>Set</code> if none required.
     * 
     * @see BindingFeature#getExtensionType()
     * @see EndpointReference#getDefaultBindingModules()
     */
    public final Set<String> getRequiredBindingFeatures() {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_getRequiredBindingFeatures();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Allows <code>BindingRT</code> to verify that it is capable of supporting Endpoint
     * policy. This MAY be important in situations where policy configures a feature which is
     * 'natively implemented' by <code>Binding</code>.
     * 
     * @param endpointReference
     *        within which natively implemented <code>BindingModule</code> instances, if any,
     *        are nested.
     * @throws DeploymentException
     *         which MUST be thrown if <code>Binding</code> cannot support all
     *         {@link PolicyExpression} configurations defined on all supported versions of
     *         specification defined on module.
     * 
     * @see BindingModule#getSupportedVersions()
     */
    public final void validateEndpointReference(EndpointReference endpointReference) throws DeploymentException {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_validateEndpointReference(endpointReference);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Sends action to referenced <code>Endpoint</code>. If referenced Service upon which
     * action is defined is 'partner' Service, action represents a request, i.e. direction is
     * IN. If referenced Service upon which action is defined is 'my' Service, action
     * represents a response, i.e. message direction is OUT and MEP is either 'in-out' or
     * 'robust-in-only.
     * 
     * @param ActionContext
     *        Defines context within which action occurs. If associated message direction is
     *        IN, action MUST be an action declared on Interface implemented by 'partner'
     *        Service, i.e. manifested as an instance of {@link MessageContext}.
     *        <p>
     *        If associated message direction is OUT and action is an output message declared
     *        on Interface implemented by 'my' Service, action MUST be manifested as an
     *        instance of {@link MessageContext}.
     *        <p>
     *        If associated message direction is OUT and action is a fault message declared on
     *        Interface implemented by 'my' Service, action MUST be manifested as an instance
     *        of {@link FaultContext} which contains a {@link BusinessFault#getMessage()}.
     *        <p>
     *        If associated message direction is OUT and action is an undeclared fault message
     *        returned by 'my' service, action MUST be manifested as an instance of
     *        {@link FaultContext} which contains a {@link SystemFault#getMessage()}.
     * 
     * @throws SystemFault
     *         Which MUST be thrown by a compliant binding if the action was NOT 'accepted for
     *         processing' by the 'receiver.' Fault code MUST be set to
     *         {@link SystemFault#DEFAULT_CODE}. Note that the rules used by binding to
     *         determine whether or not an action was 'accepted for processing' are protocol
     *         specific and MUST be defined by concrete instances of <code>BindingRT</code>.
     *         <p>
     *         Note that a compliant binding MUST NOT propogate unchecked exceptions, i.e.
     *         instances of {@link RuntimeException} which MAY occur while sending an action.
     *         Unchecked exceptions MUST be converted into an instance of
     *         <code>SystemFault</code> and handled according to the rules defined by
     *         concrete instances of <code>BindingRT</code>.
     */
    public abstract void sendAction(EndpointReference endpointReference, ActionContext actionContext) throws SystemFault;

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getName()
     */
    public final ComponentName getName() {
        return componentName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getComponentType()
     */
    public final ComponentType getComponentType() {
        return ComponentType.BINDING;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getComponentContext()
     */
    public final ComponentContext getComponentContext() {
        return componentContext;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        return fragmentIdentifier;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#deploy(org.bluestemsoftware.specification.eoa.component.ComponentContext)
     */
    public final synchronized void deploy(ComponentContext componentContext) throws DeploymentException {
        if (isDeployed == null) {
            throw new DeploymentException("Runtime component "
                    + fragmentIdentifier
                    + " is circularly referenced by "
                    + CurrentReference.getCurrentReference());
        }
        if (isDeployed) {
            return;
        }
        isDeployed = null;
        this.componentContext = componentContext;
        validate();
        isDeployed = Boolean.TRUE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#isDeployed()
     */
    public final synchronized boolean isDeployed() {
        return isDeployed;
    }

    /*
     * subclasses may override, but must call this logic
     */
    protected void validate() throws DeploymentException {
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("bindingName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(componentName);
        return sb.toString();
    }

}