/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.util.Collections;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.ext.RootExtension;
import org.bluestemsoftware.specification.eoa.ext.RootExtension.AbstractRootExtension;
import org.bluestemsoftware.specification.eoa.system.server.Feature;

/**
 * Implements a {@link Feature} on a per-message basis within 'application layer', i.e.
 * messages are instances of an abstract IDL message and are not formatted according to any
 * particular protocol. <code>ApplicationModule</code> instances are executed by
 * <code>Engine</code> when sending/receiving an action.
 * <p>
 * Examples of Features which MAY be implemented via an <code>ApplicationModule</code>
 * include validation, transformation, auditing, etc ...
 * 
 */
public abstract class ApplicationModule extends AbstractRootExtension implements FeatureModule {

    protected FragmentIdentifier fragmentIdentifier;
    protected EndpointActionReference parent;
    protected Feature feature;
    protected Short relativeOrder;
    protected Map<String, WSSpecification> wsSpecifications;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends RootExtension.Provider {

        public short spi_getRelativeOrder();
        
        public Map<String, WSSpecification> spi_getSupportedVersions();
        
        public void spi_processMessage(Message message);

    }

    public ApplicationModule(Provider provider, Feature feature) {
        super(feature.getExtensionFactory(), provider);
        this.feature = feature;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule#getFeature()
     */
    public Feature getFeature() {
        return feature;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule#getSupportedVersions()
     */
    public Map<String, WSSpecification> getSupportedVersions() {
        if (wsSpecifications == null) {
            if (provider == null) {
                throw new IllegalStateException("provider not set.");
            }
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                wsSpecifications = ((Provider)provider).spi_getSupportedVersions();
                wsSpecifications = Collections.unmodifiableMap(wsSpecifications);
            } finally {
                thread.setContextClassLoader(cl);
            }
        }
        return wsSpecifications;
    }

    /**
     * Implements <code>Feature</code> by processing abstract message instance.
     * 
     * @param message
     *        runtime instance of <code>InterfaceMessage</code> which can be retrieved via
     *        {@link Message#getMetadata()}.
     */
    public void processMessage(Message message) {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_processMessage(message);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Identifies order in which <code>ApplicationModule</code> is invoked relative to other
     * <code>ApplicationModule</code> instances defined on {@link EndpointActionReference}
     * when processing a message with direction implied by action.
     * <p>
     * Note that <code>ApplicationModule</code> processing is ordered from least to greatest.
     * For example if a module is to be the first module within the processing chain, the
     * relative value {@link Short#MIN_VALUE} SHOULD be returned. If the module is to be the
     * last module within the processing chain, the relative value {@link Short#MAX_VALUE}
     * SHOULD be returned. If processing order is irrelevant, the relative value 0 SHOULD be
     * returned.
     * <p>
     * A compliant implementation MUST provide a default value which MAY be overridden, i.e.
     * via private policy.
     * 
     * @return short relative order
     */
    public short getRelativeOrder() {
        if (relativeOrder == null) {
            if (provider == null) {
                throw new IllegalStateException("provider not set.");
            }
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                relativeOrder = Short.valueOf(((Provider)provider).spi_getRelativeOrder());
            } finally {
                thread.setContextClassLoader(cl);
            } 
        }
        return relativeOrder;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getParent()
     */
    public EndpointActionReference getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public EngineRT getRootComponent() {
        return parent.getRootComponent();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(getExtensionType());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "applicationModule",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }
    
    void setParent(EndpointActionReference parent) {
        this.parent = parent;
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("ref: ");
        sb.append(getExtensionType());
        return sb.toString();
    }

}