/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.util.Collections;
import java.util.Comparator;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.ext.RootExtension;
import org.bluestemsoftware.specification.eoa.ext.RootExtension.AbstractRootExtension;
import org.bluestemsoftware.specification.eoa.system.server.BindingFeature;
import org.bluestemsoftware.specification.eoa.system.server.Feature;

/**
 * Implements a {@link Feature} on a per-message basis within 'binding layer', i.e. messages
 * are formatted according to a specific messaging protocol, e.g. SOAP, etc ...
 * <code>BindingModule</code> instances are executed by <code>Binding</code> when
 * sending/receiving an action.
 * <p>
 * Examples of Features which MAY be implemented via a BindingModule include QOS related
 * features, e.g. ws-security, ws-transaction, etc ...
 */

public abstract class BindingModule extends AbstractRootExtension implements FeatureModule {

    protected FragmentIdentifier fragmentIdentifier;
    protected EndpointActionReference parent;
    protected BindingFeature feature;
    protected Short relativeOrder;
    protected String bindingType;
    protected Map<String, WSSpecification> wsSpecifications;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends RootExtension.Provider {

        public short spi_getRelativeOrder();

        public Map<String, WSSpecification> spi_getSupportedVersions();

        public MessageModule spi_handleMyServiceRequest(MessageModules requestModules, Content message) throws BindingFaultRT;

        public MessageModule spi_handleMyServiceResponse(MessageModules requestModules, Content message) throws BindingFaultRT;

        public MessageModule spi_handlePartnerServiceRequest(MessageModules requestModules, Content message) throws BindingFaultRT;

        public MessageModule spi_handlePartnerServiceResponse(MessageModules requestModules, MessageModules responseModules, Content message) throws BindingFaultRT;

    }

    /**
     * Constructs an instance of <code>BindingModule</code>.
     * @param provider
     * @param feature
     * @param relativeOrder
     * @param bindingType
     * @param wsSpecifications
     */

    public BindingModule(Provider provider, BindingFeature feature) {
        super(feature.getExtensionFactory(), provider);
        this.feature = feature;
        this.bindingType = feature.getBindingType();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule#getFeature()
     */
    public Feature getFeature() {
        return feature;
    }

    /**
     * Returns ref URI which identifies type of <code>Binding</code> this module's behavior
     * is intended to complement, i.e. indicates messaging protocol 'understood' by module.
     * <p>
     * Note that returned value is value defined on underlying feature, i.e. value retrieved
     * via {@link BindingFeature#getBindingType()}.
     * 
     * @return
     */
    public String getBindingType() {
        return bindingType;
    }

    /**
     * Identifies order in which <code>BindingModule</code> is invoked relative to other
     * <code>BindingModule</code> instances defined on parent {@link EndpointActionReference}
     * when processing a message with direction implied by parent action.
     * <p>
     * Note that <code>BindingModule</code> processing is ordered from least to greatest. For
     * example if a module is to be the first module within the processing chain, the relative
     * value {@link Short#MIN_VALUE} SHOULD be returned. If the module is to be the last module
     * within the processing chain, the relative value {@link Short#MAX_VALUE} SHOULD be
     * returned. If processing order is irrelevant, the relative value 0 SHOULD be returned.
     * <p>
     * A compliant implementation MUST provide a default value which MAY be overridden, i.e.
     * via private policy.
     * 
     * @return short relative order
     */
    public short getRelativeOrder() {
        if (relativeOrder == null) {
            if (provider == null) {
                throw new IllegalStateException("provider not set.");
            }
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                relativeOrder = Short.valueOf(((Provider)provider).spi_getRelativeOrder());
            } finally {
                thread.setContextClassLoader(cl);
            }
        }
        return relativeOrder;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule#getSupportedVersions()
     */
    public Map<String, WSSpecification> getSupportedVersions() {
        if (wsSpecifications == null) {
            if (provider == null) {
                throw new IllegalStateException("provider not set.");
            }
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                wsSpecifications = ((Provider)provider).spi_getSupportedVersions();
                wsSpecifications = Collections.unmodifiableMap(wsSpecifications);
            } finally {
                thread.setContextClassLoader(cl);
            }
        }
        return wsSpecifications;
    }

    /**
     * Handles a request addressed to an <code>Endpoint</code> defined on a locally hosted
     * <code>Service</code>, i.e. {@link ActorReference#isMyService()} is true and message
     * direction is 'In'.
     * <p>
     * Note that <code>BindingModule</code> SHOULD verify that sender complied with relevant
     * policy by verifying policy retrieved via {@link #getSupportedVersions()} against policy
     * manifested within message instance.
     * <p>
     * Note that if underlying feature is natively implemented by <code>Binding</code>, as
     * indicated by {@link BindingFeature#isNative()}, this method is a no-op.
     * 
     * @param requestModules
     *        Contains <code>MessageModule</code> artifacts extracted from request by
     *        <code>BindingModules</code> which were executed prior to this module.
     * 
     * @param message
     *        Root element of a concrete message instance.
     * 
     * @return messageModule messageModule decoded from request or <code>null</code> if no
     *         artifact generated as a result of processing message.
     * 
     * @throws BindingFaultRT
     *         if an error occurs while handling message
     */
    public MessageModule handleMyServiceRequest(MessageModules requestModules, Content message) throws BindingFaultRT {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        if (feature.isNative()) {
            return null;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_handleMyServiceRequest(requestModules, message);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Handles a response returned from an <code>Endpoint</code> defined on a locally hosted
     * <code>Service</code>, i.e. {@link ActorReference#isMyService()} is true and message
     * direction is 'Out'.
     * <p>
     * Note that if underlying feature is natively implemented by <code>Binding</code>, as
     * indicated by {@link BindingFeature#isNative()}, this method is a no-op.
     * 
     * @param requestModules
     *        message modules from correllated request which MAY be required to process
     *        response. Note that version of {@link WSSpecification} used to process response
     *        is implied by the version used to process correllated request.
     * 
     * @param message
     *        Root element of a concrete message instance.
     * 
     * @return messageModule messageModule encoded into response or <code>null</code> if no
     *         artifact generated as a result of processing message.
     * 
     * @throws BindingFaultRT
     *         if an error occurs while handling message
     * 
     * @see MessageModule#getVersion()
     */
    public MessageModule handleMyServiceResponse(MessageModules requestModules, Content message) throws BindingFaultRT {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        if (feature.isNative()) {
            return null;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_handleMyServiceResponse(requestModules, message);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Handles a request which is being to sent to an <code>Endpoint</code> defined on a
     * partner service, i.e. {@link ActorReference#isMyService()} is false and message
     * direction is 'In'.
     * <p>
     * Note that if underlying feature is natively implemented by <code>Binding</code>, as
     * indicated by {@link BindingFeature#isNative()}, this method is a no-op.
     * 
     * @param requestModules
     *        message modules, i.e. abstract message properties which were encoded into
     *        request. Values MAY be required to process a correllated response.
     * 
     * @param message
     *        root element of concrete message instance into which abstract message properties
     *        are encoded
     * 
     * @return messageModule messageModule encoded into request or <code>null</code> if no
     *         module encoded into request.
     * 
     * @throws BindingFaultRT
     *         if an error occurs while handling message
     */
    public MessageModule handlePartnerServiceRequest(MessageModules requestModules, Content message) throws BindingFaultRT {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        if (feature.isNative()) {
            return null;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_handlePartnerServiceRequest(requestModules, message);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Handles a response returned from an <code>Endpoint</code> defined on a partner
     * <code>Service</code>, i.e. {@link ActorReference#isMyService()} is false and message
     * direction is 'Out'.
     * <p>
     * Note that <code>BindingModule</code> SHOULD verify that sender complied with relevant
     * policy by verifying policy retrieved via {@link #getSupportedVersions()} against policy
     * manifested within message instance.
     * <p>
     * Note that if underlying feature is natively implemented by <code>Binding</code>, as
     * indicated by {@link BindingFeature#isNative()}, this method is a no-op.
     * 
     * @param requestModules
     *        message modules from correllated request which MAY be required to process
     *        response.
     * @param responseModules
     *        Contains <code>MessageModule</code> artifacts extracted by
     *        <code>BindingModules</code> which MAY have executed prior to this module.
     * 
     * @return messageModule decoded from response or <code>null</code> if no artifact
     *         generated as a result of processing message.
     * 
     * @param message
     *        Root element of a concrete message instance.
     */
    public MessageModule handlePartnerServiceResponse(MessageModules requestModules, MessageModules responseModules, Content message) {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        if (feature.isNative()) {
            return null;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_handlePartnerServiceResponse(requestModules, responseModules, message);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getParent()
     */
    public EndpointActionReference getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public EngineRT getRootComponent() {
        return parent.getRootComponent();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(getExtensionType());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "bindingModule",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(EndpointActionReference parent) {
        this.parent = parent;
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("typeRef: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(getExtensionType());
        return sb.toString();
    }

    /**
     * Orders <code>BindingModule</code> components according to relative order. Note: this
     * comparator imposes orderings that are inconsistent with equals.
     */
    public static class RelativeOrderBindingModuleComparator implements Comparator<BindingModule> {
        public int compare(BindingModule o1, BindingModule o2) {
            return o1.getRelativeOrder() - o2.getRelativeOrder();
        }
    }

}