/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.util.Collections;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.binding.UnderlyingProtocol;
import org.bluestemsoftware.specification.eoa.ext.RootExtension;
import org.bluestemsoftware.specification.eoa.ext.RootExtension.AbstractRootExtension;
import org.bluestemsoftware.specification.eoa.system.server.Feature;
import org.bluestemsoftware.specification.eoa.system.server.TransportFeature;

/**
 * Transfers messages into/out of binding layer of referenced <code>Endpoint</code> using a
 * type specific 'transport' protocol, e.g. HTTP, JMS, SMTP, etc ....
 * <p>
 * Depending upon direction, module is either the first or last feature module to process a
 * message, i.e. if message direction is 'In' module MUST be the first module in the chain, if
 * direction is 'Out', module MUST be the last module in the chain.
 * <p>
 * Type specific methods are accessed by casting <code>TransportModule</code> to a well-known
 * implementation.
 * 
 * @see UnderlyingProtocol
 */
public abstract class TransportModule extends AbstractRootExtension implements FeatureModule {

    protected FragmentIdentifier fragmentIdentifier;
    protected Feature feature;
    protected EndpointActionReference parent;
    protected Map<String, WSSpecification> wsSpecifications;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends RootExtension.Provider {

        public Map<String, WSSpecification> spi_getSupportedVersions();

    }

    public TransportModule(Provider provider, TransportFeature feature) {
        super(feature.getExtensionFactory(), provider);
        this.feature = feature;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule#getFeature()
     */
    public Feature getFeature() {
        return feature;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public EngineRT getRootComponent() {
        if (parent == null) {
            return null;
        }
        return parent.getRootComponent();
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule#getSupportedVersions()
     */
    public Map<String, WSSpecification> getSupportedVersions() {
        if (wsSpecifications == null) {
            if (provider == null) {
                throw new IllegalStateException("provider not set.");
            }
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                wsSpecifications = ((Provider)provider).spi_getSupportedVersions();
                wsSpecifications = Collections.unmodifiableMap(wsSpecifications);
            } finally {
                thread.setContextClassLoader(cl);
            }
        }
        return wsSpecifications;
    }

    /**
     * Returns parent within which this runtime component is nested.
     * 
     * @return parent or <code>null</code> if undefined, i.e. if module is default transport
     *         module instance.
     * 
     * @see TransportFeature#createDefaultTransportModule()
     */
    public EndpointActionReference getParent() {
        return parent;
    }

    void setParent(EndpointActionReference parent) {
        this.parent = parent;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(getExtensionType());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "transportModule", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("typeRef: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(getExtensionType());
        return sb.toString();
    }

}
