/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.intrface;

import org.bluestemsoftware.specification.eoa.component.Component;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;

/**
 * Generalizes InterfaceMessageReference and InterfaceFaultReference. An 'action' identifies a
 * specific message in a message exchange.
 */
public interface InterfaceAction extends Component {

    /**
     * Enumerated values which indicate whether message associated with action is 'received
     * from' a node (direction is 'In') or 'sent to' a node (direction is 'Out') where 'from'
     * and 'to' are from the perspective of Service(s) which implement parent interface.
     */
    public static enum Direction {
        IN, OUT;
    }

    /**
     * Indicates whether associated message is 'received from' a node (direction is 'In') or
     * 'sent to' a node (direction is 'Out') where 'from' and 'to' are from the perspective of
     * Service(s) which implement parent interface.
     * 
     * @return Returns enumerated value indicating message direction
     */
    public Direction getDirection();

    /**
     * Returns URI which uniquely identifes this action. URI MUST be unique within the context
     * of parent Interface component. Action URI SHOULD be globally unique.
     * <p>
     * URI is defined using rules defined within WS-Addressing 1.0 specification which MUST be
     * adhered to by an 'EOA Compliant' Server implementation.
     * 
     * @return A URI in String format which uniquely identifies referenced action
     * 
     * @see Interface#getInterfaceAction(String)
     */
    public String getAction();

    /**
     * Models a 'request' action, i.e. an action whose direction is 'In'.
     */
    public interface RequestAction extends InterfaceAction {

        /**
         * Creates a runtime instance of referenced {@link InterfaceAction}. Note that the
         * following methodology is used to generate the associated runtime
         * <code>Message</code> instance:
         * <p>
         * If a provider definition with matching name scoped to runtime
         * <code>Application</code> component's deployment is found, it is cloned and used.
         * If not, default provider defined on corresponding abstract message is cloned and
         * used.
         * 
         * @param application
         *        runtime application that is sending/receiving the action instance. If action
         *        is defined on a partner application, the application parameter represents the
         *        'sender'. If action is defined on 'my' application, the application parameter
         *        represents the 'receiver', i.e. the action is created by <code>Binding</code>.
         * 
         * @return actionContext
         * @see InterfaceMessage#getRuntimeProvider()
         * @see Message#cloneDefinition()
         */

        public ActionContext createAction(ApplicationRT application);

    }

    /**
     * Models a 'response' action, i.e. an action whose direction is 'Out'.
     */
    public interface ResponseAction extends InterfaceAction {

        /**
         * Creates a runtime instance of referenced {@link InterfaceAction}. Note that the
         * following methodology is used to generate the associated runtime
         * <code>Message</code> instance:
         * <p>
         * If a provider definition with matching name scoped to runtime
         * <code>Application</code> component's deployment is found, it is cloned and used.
         * If not, default provider defined on corresponding abstract message is cloned and
         * used.
         * 
         * @param application
         *        runtime application that is sending/receiving the action instance. If action
         *        is defined on 'my' application, the application parameter represents the
         *        'sender'. If action is defined on a partner application, the application
         *        parameter represents the 'receiver', i.e. the action is created by
         *        <code>Binding</code>.
         * @param relatesTo
         *        message id of the message being replied to.
         * 
         * @return actionContext
         * @see ApplicationRT#getDefaultMessageType()
         * @see InterfaceMessage#getRuntimeProvider()
         * @see Message#cloneDefinition()
         */
        public ActionContext createAction(ApplicationRT application, String relatesTo);

    }

}