/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.intrface;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.ResponseAction;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.BusinessFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.FaultContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSAMessageModule;

/**
 * Associates an abstract message definition defined on an InterfaceFault component with a
 * 'fault' message exchanged by parent operation.
 */
public final class InterfaceFaultReference extends NestedComponent<Interface, InterfaceOperation> implements ResponseAction, Cloneable {

    private QName faultName;
    private String action;
    private Direction direction;
    private ElementPolicy elementPolicy;

    /**
     * Constructs an <code>InterfaceFaultReference</code> instance.
     * 
     * @param faultName
     * @param action
     *        A URI which uniquely identifes this action. URI MUST be constructed using rules
     *        defined within WS-Addressing 1.0 specification. URI MUST be unique within the
     *        context of parent Interface component. URI SHOULD be globally unique.
     * @param direction
     * @param elementPolicy
     */
    public InterfaceFaultReference(QName faultName, String action, Direction direction,
            ElementPolicy elementPolicy) {
        super();
        if (faultName == null) {
            throw new IllegalArgumentException("faultName null");
        }
        if (action == null) {
            throw new IllegalArgumentException("action null");
        }
        if (direction == null) {
            throw new IllegalArgumentException("direction null");
        }
        this.faultName = faultName;
        this.action = action;
        this.direction = direction;
        this.elementPolicy = elementPolicy;
    }

    /**
     * Returns metadata describing an abstract fault declared on interface and thrown by parent
     * operation.
     * <p>
     * 
     * @return InterfaceFault object
     */
    public final InterfaceFault getFault() {
        return ((Interface)getRootComponent()).getFault(faultName);
    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * 
     * @return ElementPolicy or <code>null</code> if no policy attached.
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction#getAction()
     */
    public String getAction() {
        return action;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction#getDirection()
     */
    public Direction getDirection() {
        return direction;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.ResponseAction#createAction(org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT, java.lang.String)
     */
    public ActionContext createAction(ApplicationRT application, String relatesTo) {
        if (relatesTo == null || relatesTo.equals("")) {
            throw new IllegalArgumentException("Failed to create response action."
                    + " RelatesTo messageID null or empty string");
        }
        WSAMessageModule wsamm = new WSAMessageModule(action);
        wsamm.setRelatesTo(relatesTo);
        MessageModules mm = new MessageModules(wsamm);
        QName faultName = null;
        InterfaceMessage am = null;
            InterfaceFault interfaceFault = getFault();
            faultName = interfaceFault.getName();
            am = interfaceFault.getReferencedComponent();
        ComponentDeployment deployment = application.getComponentContext().getDeployment();
        Message m = (Message)deployment.getProvider(ComponentType.MESSAGE, am.getName());
        if (m == null) {
            m = ((Message)am.getRuntimeProvider()).cloneDefinition();
        } else {
            m = m.cloneDefinition();
        }
        return new FaultContext(mm, new BusinessFault(faultName, m));
    }
    
    final void validate() throws DeploymentException {
        // TODO: implement
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     * 
     * Note that we need to account for fact that referenced fault may be inherited, i.e. ns
     * is different than ns of parent interface.
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(((InterfaceOperation)parent).getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append("Out");
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(faultName.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + faultName.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "interfaceFaultReference", pointerPart.toString());
        }
        return fragmentIdentifier;
    }
    
    @Override
    public Object clone() throws CloneNotSupportedException {
        InterfaceFaultReference ifr =  (InterfaceFaultReference)super.clone();
        if (elementPolicy != null) {
            ifr.elementPolicy = (ElementPolicy)elementPolicy.clone();
        }
        ifr.parent = null;
        return ifr;
    }

    void setParent(InterfaceOperation parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("faultName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(faultName);
        sb.append(System.getProperty("line.separator"));
        sb.append("action: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(action);
        sb.append(System.getProperty("line.separator"));
        sb.append("direction: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(direction);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(faultName) + hashCode(action) + hashCode(direction);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof InterfaceFaultReference)) {
            return false;
        }

        if ((faultName == ((InterfaceFaultReference)obj).faultName || (faultName != null && faultName
                .equals(((InterfaceFaultReference)obj).faultName)))
                && (action == ((InterfaceFaultReference)obj).action || (action != null && action
                        .equals(((InterfaceFaultReference)obj).action)))
                && (direction == ((InterfaceFaultReference)obj).direction || (direction != null && direction
                        .equals(((InterfaceFaultReference)obj).direction)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}