/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.intrface;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.types.ElementDeclaration;

/**
 * Models an abstract operation. Operation defines an input action, an optional output action
 * and optional fault actions.
 */

public final class InterfaceOperation extends NestedComponent<Interface, Interface> implements Cloneable {

    private QName name;
    private MEP mep;
    private Set<String> styles;
    private boolean isSafe;
    private InterfaceMessageReference inputMessageReference;
    private InterfaceMessageReference outputMessageReference;
    private Map<QName, InterfaceFaultReference> faultReferences;
    private ElementPolicy elementPolicy;

    public InterfaceOperation(QName operationName, MEP mep, Set<String> styles, boolean isSafe,
            InterfaceMessageReference inputMessageReference, InterfaceMessageReference outputMessageReference,
            Map<QName, InterfaceFaultReference> faultReferences, ElementPolicy elementPolicy) {
        super();
        if (operationName == null) {
            throw new IllegalArgumentException("operationName null");
        } else {
            this.name = operationName;
        }
        if (mep == null) {
            throw new IllegalArgumentException("mep null");
        } else {
            this.mep = mep;
        }
        if (styles == null) {
            throw new IllegalArgumentException("styles null");
        } else {
            this.styles = styles;
        }
        this.isSafe = isSafe;
        this.inputMessageReference = inputMessageReference;
        this.outputMessageReference = outputMessageReference;
        if (faultReferences == null) {
            throw new IllegalArgumentException("faultReferences null");
        } else {
            this.faultReferences = faultReferences;
        }
        if (elementPolicy != null) {
            this.elementPolicy = elementPolicy;
        }
    }

    /**
     * Models an operation <code>Style</code>. Type specific methods are accessed by casting
     * <code>Style</code> to a well-known implementation.
     * 
     * @see ElementDeclaration#toStyle(String)
     */
    public interface Style {

        /**
         * Gets IRI which uniquely identifies style.
         * 
         * @return IRI formatted as a String
         */
        public String getStyleIRI();

    }

    /**
     * Enumeration defining supported message exchange patterns. Note that the pattern
     * robust-in-only is NOT backwards compatible with WSDL 1.1. If used, the operation cannot
     * be described via WSDL 1.1.
     */
    public static enum MEP {
        IN_ONLY, ROBUST_IN_ONLY, IN_OUT
    }

    /**
     * Gets qualified name of interface operation. Namespace is same as parent interface unless
     * operation was inherited from an interface extended by parent interface.
     * <p>
     * 
     * @return Qualifed name of operation.
     */
    public final QName getName() {
        return name;
    }

    /**
     * Retrieves 'message exchange pattern' utilized by this operation. A MEP defines a
     * sequence of 'placeholder messages' all of which are are related to a single 'exchange.'
     * 
     * @return Enumerated value indicating a predefined message exchange pattern
     */
    public final MEP getMessageExchangePattern() {
        return mep;
    }

    /**
     * Defines operation style which MAY provide additional constraints on the element
     * declarations used by the operation.
     * <p>
     * Note that this attribute is NOT backwards compatible with WSDL 1.1. If defined, this
     * attribute cannot be described via WSDL 1.1.
     * 
     * @return a Set of style iri's or an empty Set if style is undefined.
     */
    public Set<String> getStyles() {
        return styles;
    }

    /**
     * Indicates whether the operation is asserted to be safe for users to invoke, e.g. a 'read
     * only' operation.
     * 
     * @param operationName
     * @return true if safe, false if no assertion can be made regarding operation safety.
     * 
     * @see http://www.w3.org/TR/wsdl20-adjuncts/#safety
     */
    public boolean isSafe() {
        return isSafe;
    }

    /**
     * Retrieves message reference object whose message direction is 'In'. The returned object
     * refers to an 'ordinary' abstract message.
     * 
     * @return interfaceMessageReference
     */
    public final RequestMessageReference getInputMessageReference() {
        return (RequestMessageReference)inputMessageReference;
    }

    /**
     * Retrieves message reference object whose message direction is 'Out'. The returned object
     * refers to an 'ordinary' abstract message.
     * 
     * @return interfaceMessageReference or null if message exchange pattern is IN_ONLY.
     */
    public final ResponseMessageReference getOutputMessageReference() {
        return (ResponseMessageReference)outputMessageReference;
    }

    /**
     * Retrieves reference to an abstract fault declared on parent interface and thrown by this
     * operation.
     * <p>
     * @param QName
     *        Qualified name of fault. Note that unless the fault was inherited from an extended
     *        interface, the namespace is same as namespace within which parent interface is
     *        defined.
     * 
     * @return InterfaceFaultReference object, or null if no fault found matching faultName or
     *         operation mep is IN_ONLY.
     */
    public final InterfaceFaultReference getFaultReference(QName faultName) {
        if (faultReferences != null) {
            return faultReferences.get(faultName);
        }
        return null;
    }

    /**
     * Retrieves all fault reference objects each of which contains an abstract fault declared
     * thrown by this operation.
     * <p>
     * 
     * @return Collection of InterfaceFaultReference objects. If no faults are defined for this
     *         operation or if message exchange pattern is IN_ONLY, an empty Collection is
     *         returned.
     */
    @SuppressWarnings("unchecked")
    public final Collection<InterfaceFaultReference> getFaultReferences() {
        if (faultReferences != null) {
            return Collections.unmodifiableCollection(faultReferences.values());
        } else {
            return Collections.EMPTY_LIST;
        }

    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * 
     * @return ElementPolicy or <code>null</code> if no policy attached.
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    final void validate() throws DeploymentException {
        if (inputMessageReference != null) {
            inputMessageReference.validate();
        }
        if (outputMessageReference != null) {
            outputMessageReference.validate();
        }
        if (faultReferences != null) {
            for (InterfaceFaultReference faultReference : faultReferences.values()) {
                faultReference.validate();
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     * 
     * Note that we must account for fact that component may be inherited, i.e. namespace may
     * be different than namespace of parent interface.
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(rootName.getLocalPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(name.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + name.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "interfaceOperation", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(Interface parent) {
        this.parent = parent;
        if (inputMessageReference != null) {
            inputMessageReference.setParent(this);
        }
        if (outputMessageReference != null) {
            outputMessageReference.setParent(this);
        }
        if (faultReferences != null) {
            for (InterfaceFaultReference faultReference : faultReferences.values()) {
                faultReference.setParent(this);
            }
        }
    }

    final void establishReference() throws DeploymentException {
        if (inputMessageReference != null) {
            inputMessageReference.establishReference();
        }
        if (outputMessageReference != null) {
            outputMessageReference.establishReference();
        }
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#clone()
     */
    @Override
    public Object clone() throws CloneNotSupportedException {
        InterfaceOperation io = (InterfaceOperation)super.clone();
        if (faultReferences.size() > 0) {
            io.faultReferences = new HashMap<QName, InterfaceFaultReference>();
            InterfaceFaultReference clone = null;
            for (InterfaceFaultReference temp : faultReferences.values()) {
                clone = (InterfaceFaultReference)temp.clone();
                io.faultReferences.put(clone.getFault().getName(), clone);
            }
        }
        InterfaceMessageReference clone = null;
        if (inputMessageReference != null) {
            clone = (InterfaceMessageReference)inputMessageReference.clone();
            io.inputMessageReference = clone;
        }
        if (outputMessageReference != null) {
            clone = (InterfaceMessageReference)outputMessageReference.clone();
            io.outputMessageReference = clone;
        }
        if (elementPolicy != null) {
            io.elementPolicy = (ElementPolicy)elementPolicy.clone();
        }
        io.parent = null;
        return io;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("mep: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(mep);
        sb.append(System.getProperty("line.separator"));
        sb.append("styles: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(styles);
        sb.append(System.getProperty("line.separator"));
        sb.append("isSafe: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(isSafe);
        sb.append(System.getProperty("line.separator"));
        sb.append("inputMessageReference: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(inputMessageReference);
        sb.append(System.getProperty("line.separator"));
        sb.append("outputMessageReference: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(outputMessageReference);
        sb.append(System.getProperty("line.separator"));
        sb.append("faultReferences: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(faultReferences);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(name)
                + hashCode(mep)
                + hashCode(styles)
                + hashCode(isSafe)
                + hashCode(inputMessageReference)
                + hashCode(outputMessageReference)
                + hashCode(faultReferences);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof InterfaceOperation)) {
            return false;
        }

        if ((name == ((InterfaceOperation)obj).name || (name != null && name
                .equals(((InterfaceOperation)obj).name)))
                && (mep == ((InterfaceOperation)obj).mep || (mep != null && mep
                        .equals(((InterfaceOperation)obj).mep)))
                && (styles == ((InterfaceOperation)obj).styles || (styles != null && styles
                        .equals(((InterfaceOperation)obj).styles)))
                && (isSafe == ((InterfaceOperation)obj).isSafe)
                && (inputMessageReference == ((InterfaceOperation)obj).inputMessageReference || (inputMessageReference != null && inputMessageReference
                        .equals(((InterfaceOperation)obj).inputMessageReference)))
                && (outputMessageReference == ((InterfaceOperation)obj).outputMessageReference || (outputMessageReference != null && outputMessageReference
                        .equals(((InterfaceOperation)obj).outputMessageReference)))
                && (faultReferences == ((InterfaceOperation)obj).faultReferences || (faultReferences != null && faultReferences
                        .equals(((InterfaceOperation)obj).faultReferences)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}