/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.intrface.rt;

import java.util.HashMap;

import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference.ContentModel;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationDataModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSAMessageModule;

/**
 * Models a runtime instance of {@link InterfaceAction}. Defines 'context' within which action
 * occurs.
 */

public abstract class ActionContext {

    protected MessageModules messageModules;
    protected Message message;

    public ActionContext() {
    }

    /**
     * Constructs an instanceof <code>ActionContext</code>.
     * 
     * @param messageModules
     *        which MUST contain an instanceof {@link WSAMessageModule}
     * @param message
     */
    public ActionContext(MessageModules messageModules, Message message) {
        if (messageModules == null) {
            throw new IllegalArgumentException("messageModules null");
        }
        this.messageModules = messageModules;
        this.message = message;
    }

    /**
     * Convenience method which returns URI which uniquely identifes action within context of
     * referenced <code>Endpoint</code>, i.e. returns {@link WSAMessageModule#getAction()}.
     * 
     * @return A URI in String format.
     * 
     * @see InterfaceAction#getAction()
     */
    public final String getAction() {
        return messageModules.getAction();
    }

    /**
     * Convenience method which returns wsa:MessageID value defined on built-in module
     * ws-addressing, i.e. {@link WSAMessageModule#getMessageID()}.
     * 
     * @return messageID
     */
    public final String getMessageID() {
        return messageModules.getMessageID();
    }

    /**
     * Convenience method which returns wsa:RelatesTo value defined on built-in module
     * ws-addressing, i.e. {@link WSAMessageModule#getRelatesTo()}.
     * 
     * @return relatesTo
     */
    public final String getRelatesTo() {
        return messageModules.getRelatesTo();
    }

    /**
     * Convenience method used to retreive a specific message module.
     * @param type
     *        identifies module type
     * @return message module or <code>null</code> if undefined
     */
    public final <T extends MessageModule> T getMessageModule(Class<T> type) {
        return type.cast(messageModules.get(type));
    }

    /**
     * Gets message modules.
     * @return messageModules.
     */
    public final MessageModules getMessageModules() {
        return messageModules;
    }

    /**
     * Returns runtime instance of InterfaceMessage which initiated this action.
     * 
     * @return A runtime message instance or <code>null</code> if action associated with a
     *         normal message and content is modeled as {@link ContentModel#NONE}. If
     *         content is modeled as {@link ContentModel#OTHER}, the behavior is undefined.
     */
    public final Message getMessage() {
        return message;
    }

    /**
     * Contains <code>MessageModule</code> instances related to a specific action instance,
     * i.e. which is manifested as an instance of <code>ActionContext</code>.
     */
    public static class MessageModules extends HashMap<Class<?>, MessageModule> {

        private static final long serialVersionUID = 1L;

        /**
         * Constructs an instance of <code>MessageModules</code>.
         * 
         * @param wsaMessageModule
         *        Note that <code>WSAMessageModule</code> definition is built-in to this
         *        specification and as such, an instance MUST be provided for each new instance
         *        of this class.
         *        <p>
         *        Note that <code>ApplicationDataModule</code> definition is built-in to this
         *        specification and as such, an instance is created and added by default.
         */
        public MessageModules(WSAMessageModule wsaMessageModule) {
            if (wsaMessageModule == null) {
                throw new IllegalArgumentException("wsaMessageModule null");
            }
            put(WSAMessageModule.class, wsaMessageModule);
            put(ApplicationDataModule.class, new ApplicationDataModule());
        }

        /**
         * Gets <code>MessageModule</code> by module type.
         * <p>
         * Note that <code>WSAMessageModule</code> and <code>ApplicationDataModule</code>
         * definition are built-in to this specification and are always available.
         * 
         * @param <T>
         * @param type
         * @return Module or <code>null</code> if undefined. Note that
         *         <code>WSAMessageModule</code> and <code>ApplicationDataModule</code>
         *         definition are built-in to this specification and are always defined.
         */
        public <T extends MessageModule> T getModule(Class<T> type) {
            return type.cast(get(type));
        }

        /**
         * Adds a <code>MessageModule</code>. Note that if module exists of implied type, it
         * is overwritten with provided version.
         * 
         * @param messageModule
         */
        public void putModule(MessageModule messageModule) {
            if (messageModule == null) {
                throw new IllegalArgumentException("messageModule null");
            }
            put(messageModule.getClass(), messageModule);
        }

        /**
         * Convenience method which returns action value defined on built-in module
         * {@link WSMessageModule}
         * 
         * @return action URI
         */
        public String getAction() {
            return getModule(WSAMessageModule.class).getAction();
        }

        /**
         * Convenience method which returns messageID value defined on built-in module
         * {@link WSMessageModule}
         * 
         * @return messageID
         */
        public String getMessageID() {
            return getModule(WSAMessageModule.class).getMessageID();
        }

        /**
         * Convenience method which returns relatesTo value defined on built-in module
         * {@link WSMessageModule}
         * 
         * @return messageID or <code>null</code> if undefined, i.e. if this is a request
         *         action
         */
        public String getRelatesTo() {
            return getModule(WSAMessageModule.class).getRelatesTo();
        }

    }

}