/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.intrface.rt;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.bluestemsoftware.specification.eoa.component.engine.rt.BindingModule;
import org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule.WSSpecification;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * An artifact of {@link BindingModule} processing. Represents a collection of abstract message
 * properties extracted by module from a concrete message instance. Methods specific to module
 * type are invoked by casting module to a well-known implementation. Implementations SHOULD be
 * immutable.
 */

public abstract class MessageModule {

    /**
     * Returns ref URI which uniquely identifies <code>MessageModule</code> type. Value
     * returned MUST be fully qualified class name of concrete class which extends this
     * abstract class.
     * 
     * @return String ref URI
     */
    public abstract String getType();

    /**
     * Identifies version of specification which was used by <code>BindingModule</code> to
     * decode module from concrete message instance, which MAY be required to identify version
     * required to encode response.
     * 
     * @return namespace or <code>null</code> if module not extracted from a concrete message
     *         instance.
     * 
     * @see WSSpecification#getNamespace()
     */
    public abstract String getNamespace();

    /**
     * Converts this <code>MessageModule</code> instance into an <code>Element</code>.
     * Allows module to be serialized using a portable approach.
     * <p>
     * Note that <code>Element</code> MUST NOT be serialized according to any specific
     * messaging protocol, i.e. the <code>Element</code> is an abstract representation of
     * underlying feature.
     * 
     * @param ownerDocument
     * @return Element or <code>null</code> if module is empty
     * @see #valueOf(Class, Element)
     */
    public abstract Element toElement(Document ownerDocument);

    /**
     * Converts an <code>Element</code> into a type specific instanceof
     * <code>MessageModule</code>.
     * 
     * @param moduleType
     *        fully qualified name of Class which extends <code>MessageModule</code> and
     *        which MUST define static method valueOf(org.w3c.dom.Element) which returns
     *        <code>Element</code>. Note that the thread's current context classloader is
     *        used to load concrete instance of <code>MessageModule</code>.
     * 
     * @param element
     *        element instance to convert
     * @return
     * @throws InvocationTargetException
     * @throws IllegalAccessException
     * @throws
     * @see #toElement(Document)
     * @see #getType()
     */
    public static MessageModule valueOf(String moduleType, Element element) throws ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
        if (moduleType == null) {
            throw new IllegalArgumentException("moduleType null");
        }
        if (element == null) {
            throw new IllegalArgumentException("element null");
        }
        Class<?> moduleClass = Thread.currentThread().getContextClassLoader().loadClass(moduleType);
        Method method = moduleClass.getMethod("valueOf", new Class[] { Element.class });
        return (MessageModule)method.invoke(null, element);
    }

}