/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.message;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.types.Schema;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent.Type;

/**
 * Metadata describing an abstract message part, i.e. a part within a wsdl message. Provides
 * backwards compatibility with wsdl 1.1.
 */
public class MessagePart extends NestedComponent<InterfaceMessage, InterfaceMessage> implements SchemaComponentReference {

    private String partName;
    private String schemaComponentNamespace;
    private Type schemaComponentType;
    private String schemaComponentName;
    private ComponentName schemaName;
    private Schema schema;
    private SchemaComponent schemaComponent;

    /**
     * Defines default name for message 'payload' part which MUST be used when a default name
     * is required, e.g. when an <code>InterfaceMessage</code> definition is read from a WSDL
     * 2.x description.
     * 
     * @see InterfaceMessage
     */
    public static final String PAYLOAD = "payload";

    public MessagePart(String partName, String schemaComponentNamespace, Type schemaComponentType,
            String schemaComponentName) {
        super();
        if (partName == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.partName = partName;
        }
        if (schemaComponentNamespace == null) {
            throw new IllegalArgumentException("schemaComponentNamespace null");
        } else {
            this.schemaComponentNamespace = schemaComponentNamespace;
        }
        if (schemaComponentType == null) {
            throw new IllegalArgumentException("schemaComponentType null");
        } else {
            this.schemaComponentType = schemaComponentType;
        }
        if (schemaComponentName == null) {
            throw new IllegalArgumentException("schemaComponentName null");
        } else {
            this.schemaComponentName = schemaComponentName;
        }
        // note that local name of referenced schema is unknown until
        // reference is established, at which point it will be updated
        schemaName = new ComponentName(schemaComponentNamespace, "foo");
    }

    /**
     * Gets name of abstract part.
     * 
     * @return
     */
    public String getName() {
        return partName;
    }

    /**
     * Gets targetNamespace within which referenced schema component is defined.
     * @return
     */
    public String getSchemaComponentNamespace() {
        return schemaComponentNamespace;
    }

    /**
     * Convenience method which retrieves qualified name of referenced schema component.
     * @return
     */
    public QName getSchemaComponentQName() {
        return new QName(schemaComponentNamespace, schemaComponentName);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaComponentReference#getSchemaComponentName()
     */
    public String getSchemaComponentName() {
        return schemaComponentName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaComponentReference#getSchemaComponentType()
     */
    public Type getSchemaComponentType() {
        return schemaComponentType;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<Schema> getReferencedComponentClass() {
        return Schema.class;
    }

    /**
     * Gets referenced <code>SchemaComponent</code>.
     * 
     * @return SchemaComponent referenced <code>SchemaComponent</code>
     */
    public SchemaComponent getSchemaComponent() {
        return schemaComponent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public Schema getReferencedComponent() {
        return schema;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return schemaName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.SCHEMA;
    }

    void establishReference() throws DeploymentException {
        if (schemaComponent != null) {
            return;
        }
        schema = (Schema)getRootComponent().getComponentContext().findReferencedComponent(this);
        if (schema == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Schema which defines schema component "
                    + new QName(schemaComponentNamespace, schemaComponentName)
                    + " typed as a "
                    + schemaComponentType
                    + ", is undefined.");
        } else {
            schemaComponent = schema.getSchemaComponent(schemaComponentType, schemaComponentName);
            // now that localname of schema is defined update our name
            schemaName = schema.getName();
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(rootName.getLocalPart());
            pointerPart.append("/");
            pointerPart.append(partName);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "interfaceMessagePart", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(InterfaceMessage parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(getName());
        sb.append(System.getProperty("line.separator"));
        sb.append("schemaComponentName:");
        sb.append(schemaComponentName);
        sb.append(System.getProperty("line.separator"));
        sb.append("schemaComponentType:");
        sb.append(schemaComponentType);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(getName()) + hashCode(schemaComponentName) + hashCode(schemaComponentType);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof MessagePart)) {
            return false;
        }

        if ((getName() == ((MessagePart)obj).getName() || (getName() != null && getName().equals(
                ((MessagePart)obj).getName())))

        ) {
            if (schemaComponentName == ((MessagePart)obj).schemaComponentName
                    || (schemaComponentName != null && schemaComponentName
                            .equals(((MessagePart)obj).schemaComponentName))) {
                if (schemaComponentType == ((MessagePart)obj).schemaComponentType
                        || (schemaComponentType != null && schemaComponentType
                                .equals(((MessagePart)obj).schemaComponentType))) {
                    return true;
                }
            }
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}