/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.message.rt;

import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;

import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.MessageFragment;
import org.bluestemsoftware.specification.eoa.component.message.MessagePart;
import org.bluestemsoftware.specification.eoa.component.rt.ProviderReader;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT;
import org.bluestemsoftware.specification.eoa.component.types.SchemaProperty;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaException;
import org.bluestemsoftware.specification.eoa.ext.RootExtension;
import org.bluestemsoftware.specification.eoa.ext.expression.EvaluationException;
import org.bluestemsoftware.specification.eoa.ext.message.MessageFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * Models a runtime instance of {@link InterfaceMessage} which is delivered to, or returned
 * from, a runtime <code>Application</code> instance. Message instances MAY be cast to a
 * 'well-known' implementation for access to type specific methods.
 * <p>
 * Note that <code>Message</code> is not an instance of {@link SingletonRT} which implies
 * that more than one instance MAY exist for a specific instance of {@link InterfaceMessage}.
 * Allows an <code>Application</code> to 'scope' a runtime provider definition within it's
 * deployment, effectively 'hiding' the 'default' message instance associated with abstract
 * message.
 * 
 * @see ApplicationRT#createMessage(javax.xml.namespace.QName, boolean)
 */

public interface Message extends RootExtension, RootComponentRT, Document, ContentFactory {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends RootExtension.Provider {
    }

    public InterfaceMessage getMetadata();

    /**
     * Clones underlying provider definition, i.e. definition which was used to create message
     * via {@link ProviderReader#readProvider(ComponentContext, DOMSource)} or
     * {@link MessageFactory#createDefaultProvider(ComponentContext, InterfaceMessage)}.
     * <p>
     * Note that any changes to message content which affect the underlying definition MUST be
     * effectively ignored for the purposes of the clone operation.
     * 
     * @return
     */
    public Message cloneDefinition();

    /**
     * Gets message content, i.e. message 'payload'.
     * <p>
     * Note that this method is intended for use on 'WSDL 2.0 style' messages, i.e. messages
     * that define a single abstract part typed as an element.
     * 
     * @return An element which represents message content or <code>null</code> if message
     *         has no content.
     * 
     * @throws UnsupportedOperationException
     *         Which MUST be thrown if abstract message not a 'WSDL 2.0 style' message.
     */
    public Content getContent();

    /**
     * Retrieves content from a specific message part.
     * <p>
     * Note that this method is provided for backwards compatibility with WSDL 1.1 multi-part
     * messages and its use is discouraged.
     * 
     * @param partName
     *        name of abstract part.
     * 
     * @return An element which represents part content. The instance returned is dependent
     *         upon how abstract part is typed:
     *         <p>
     *         If abstract part Type is typed as ELEMENT, an instance of indicated element is
     *         returned.
     *         <p>
     *         If abstract part is typed as a SIMPLE type, a 'part accessor' element containing
     *         an instance of indicated Type with no namespace and whose local name equals part
     *         name is returned.
     *         <p>
     *         If abstract part is typed as a COMPLEX type , a 'part accessor' element, whose
     *         content is an instance of modeled content, with no namespace and whose local
     *         name equals part name is returned.
     *         <p>
     *         Or <code>null</code> is returned if part has no content.
     * @throws IllegalArgumentException
     *         Which MUST be thrown if partName fails to match a part defined on abstract
     *         message.
     */
    public Content getContent(String partName);

    /**
     * Sets message content, i.e. message 'payload'. Existing content, if any, is overwritten.
     * <p>
     * Note that this method is intended for use on 'WSDL 2.0' style messages, i.e. messages
     * that define a single abstract part typed as an element.
     * 
     * @parm content An element representing message payload or <code>null</code> in which
     *       case existing payload, if any, MUST be cleared.
     * 
     * @throws UnsupportedOperationException
     *         Which MUST be thrown if abstract message not a 'WSDL 2.0 style' message.
     */
    public void setContent(Content content);

    /**
     * Sets content on a specific message part. Existing content, if any, is overwritten.
     * <p>
     * Note that this method is provided for backwards compatibility with WSDL 1.1 multi-part
     * messages and its use is discouraged.
     * 
     * @param partName
     *        name of abstract part.
     * 
     * @parm content An element which represents part content. The instance passed is dependent
     *       upon how abstract part is typed:
     *       <p>
     *       If abstract part Type is typed as ELEMENT, an instance of indicated element is
     *       passed.
     *       <p>
     *       If abstract part is typed as a SIMPLE type, a 'part accessor' element, containing
     *       an instance of indicated Type, with no namespace and whose local name equals part
     *       name is passed.
     *       <p>
     *       If abstract part is typed as a COMPLEX type , a 'part accessor' element, whose
     *       content is an instance of modeled content, with no namespace and whose local name
     *       equals part name is passed.
     *       <p>
     *       If content is <code>null</code>, part content MUST be cleared.
     * @throws IllegalArgumentException
     *         Which MUST be thrown if partName fails to match a part defined on abstract
     *         message.
     */
    public void setContent(String partName, Content content);

    /**
     * Evaluates a message fragment expression against message part content, i.e. the part upon
     * which fragment was defined and returns a <code>String</code>.
     * <p>
     * Note that the context <code>Node</code> from which fragment expression is evaluated
     * MUST be a synthetic Document Information Item which contains a single Element
     * Information Item representing root element of message part upon which the fragment is
     * defined. This is necessary to insure portability of message fragment definitions, i.e.
     * internal structure of message document may differ by provider which could break absolute
     * query paths.
     * 
     * @param propertyName
     *        Matches name of a {@link MessageFragment} defined on abstract message which will
     *        be used to extract an instance of indicated {@link SchemaProperty} from this
     *        runtime message instance.
     * @return String Note that the rules used to perform conversion of result to type
     *         <code>String</code> are expression language specific.
     * @throws EvaluationException
     *         If an error occurs evaluating expression.
     * 
     * @see InterfaceMessage#getMessageFragments()
     * 
     */
    public String evaluateString(QName propertyName) throws EvaluationException;

    /**
     * Evaluates a message fragment expression against message part content, i.e. the part upon
     * which fragment was defined and returns a <code>Number</code>.
     * <p>
     * Note that the context <code>Node</code> from which fragment expression is evaluated
     * MUST be a synthetic Document Information Item which contains a single Element
     * Information Item representing root element of message part upon which the fragment is
     * defined. This is necessary to insure portability of message fragment definitions, i.e.
     * internal structure of message document may differ by provider which could break absolute
     * query paths.
     * 
     * @param propertyName
     *        Matches name of a {@link MessageFragment} defined on abstract message which will
     *        be used to extract an instance of indicated {@link SchemaProperty} from this
     *        runtime message instance.
     * @return Number Note that the rules used to perform conversion of result to type
     *         <code>Number</code> are expression language specific.
     * @throws EvaluationException
     *         If an error occurs evaluating expression.
     * 
     * @see InterfaceMessage#getMessageFragments()
     * 
     */
    public Number evaluateNumber(QName propertyName) throws EvaluationException;

    /**
     * Evaluates a message fragment expression against message part content, i.e. the part upon
     * which fragment was defined, and returns a <code>Boolean</code>.
     * <p>
     * Note that the context <code>Node</code> from which fragment expression is evaluated
     * MUST be a synthetic Document Information Item which contains a single Element
     * Information Item representing root element of message part upon which the fragment is
     * defined. This is necessary to insure portability of message fragment definitions, i.e.
     * internal structure of message document may differ by provider which could break absolute
     * query paths.
     * 
     * @param propertyName
     *        Matches name of a {@link MessageFragment} defined on abstract message which will
     *        be used to extract an instance of indicated {@link SchemaProperty} from this
     *        runtime message instance.
     * @return Boolean Note that the rules used to perform conversion of result to type
     *         <code>Boolean</code> are expression language specific.
     * @throws EvaluationException
     *         If an error occurs evaluating expression.
     * 
     * @see InterfaceMessage#getMessageFragments()
     * 
     */
    public Boolean evaluateBoolean(QName propertyName) throws EvaluationException;

    /**
     * Evaluates a message fragment expression against message part content, i.e. the part upon
     * which fragment was defined, and returns a <code>Node</code>.
     * <p>
     * Note that the context <code>Node</code> from which fragment expression is evaluated
     * MUST be a synthetic Document Information Item which contains a single Element
     * Information Item representing root element of message part upon which the fragment is
     * defined. This is necessary to insure portability of message fragment definitions, i.e.
     * internal structure of message document may differ by provider which could break absolute
     * query paths.
     * 
     * @param propertyName
     *        Matches name of a {@link MessageFragment} defined on abstract message which will
     *        be used to extract an instance of indicated {@link SchemaProperty} from this
     *        runtime message instance.
     * @return Node Note that the rules used to perform conversion of result to type
     *         <code>Node</code> are expression language specific.
     * @throws EvaluationException
     *         If an error occurs evaluating expression.
     * 
     * @see InterfaceMessage#getMessageFragments()
     * 
     */
    public Node evaluateNode(QName propertyName) throws EvaluationException;

    /**
     * Evaluates a message fragment expression against message part content, i.e. the part upon
     * which fragment was defined, and returns a <code>Set</code>.
     * <p>
     * Note that the context <code>Node</code> from which fragment expression is evaluated
     * MUST be a synthetic Document Information Item which contains a single Element
     * Information Item representing root element of message part upon which the fragment is
     * defined. This is necessary to insure portability of message fragment definitions, i.e.
     * internal structure of message document may differ by provider which could break absolute
     * query paths.
     * 
     * @param propertyName
     *        Matches name of a {@link MessageFragment} defined on abstract message which will
     *        be used to extract an instance of indicated {@link SchemaProperty} from this
     *        runtime message instance.
     * @return List of matching <code>Nodes</code>. Note that the rules used to perform
     *         conversion of result to type <code>List</code> are expression language
     *         specific.
     * @throws EvaluationException
     *         If an error occurs evaluating expression.
     * 
     * @see InterfaceMessage#getMessageFragments()
     * 
     */
    public List<Node> evaluateNodes(QName propertyName) throws EvaluationException;

    /**
     * Validates message, i.e. validates content of each message part, if any, against schema
     * document within which referenced schema component is defined.
     * 
     * @throws SchemaException
     * @see MessagePart#getSchemaComponent()
     */
    public void validate() throws MessageValidationException;

}
