/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.policy;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyDocument;

/**
 * A deployable, 'top-level' component which wraps a runtime <code>Policy</code> instance,
 * i.e. a <code>PolicyDocument</code>.
 */
public final class Policy extends RootComponent implements RuntimeProvidable {

    /**
     * Constructs a <code>Policy</code> component. Note that runtime provider, i.e.
     * <code>PolicyDocument</code> is set by <code>System</code> when component is deployed.
     * @param name
     *        qualified name of expression which MUST be composed as follows. The namespace URI
     *        is target namespace of IDL document within which expression is defined. Local
     *        name is expression id, which MUST NOT be <code>null</code> and which MUST NOT
     *        be empty string.
     */
    public Policy(ComponentName name) {
        super();
        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        fragmentIdentifier = new FragmentIdentifier(name.getNamespaceURI(), Scheme.EOA, "policy", name
                .getLocalPart());
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable#getRuntimeProvider()
     */
    public PolicyDocument getRuntimeProvider() {
        if (runtimeProvider == null || !runtimeProvider.isDeployed()) {
            throw new IllegalStateException("Provider definition not yet deployed.");
        }
        return (PolicyDocument)runtimeProvider;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#establishReferences()
     */
    @Override
    protected synchronized void establishReferences() throws DeploymentException {
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#getComponentType()
     */
    @Override
    public final ComponentType getComponentType() {
        return ComponentType.POLICY;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#validate()
     */
    @Override
    protected final void validate() throws DeploymentException {
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(name);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(name);
    }

    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Policy)) {
            return false;
        }

        if (getName() == ((Policy)obj).getName()
                || (getName() != null && getName().equals(((Policy)obj).getName()))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
