/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.policy.rt;

import java.util.Set;

import org.bluestemsoftware.specification.eoa.component.engine.Engine;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointOperationReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule;
import org.bluestemsoftware.specification.eoa.component.service.Endpoint;
import org.bluestemsoftware.specification.eoa.component.service.EndpointAction;
import org.bluestemsoftware.specification.eoa.component.service.EndpointOperation;
import org.bluestemsoftware.specification.eoa.component.service.Service;
import org.bluestemsoftware.specification.eoa.system.server.ApplicationFeature;
import org.bluestemsoftware.specification.eoa.system.server.TransportFeature;

/**
 * Aggregates policy expressions from all four policy subjects which MAY affect processing of
 * messages associated with a specific action. Action policy is either 'private' or 'public'
 * and is used to enable/constrain {@link FeatureModule} instances, i.e. it is used by
 * feature implemented by module to create and configure the module.
 * 
 * @see ApplicationFeature#createApplicationModule(EndpointActionReference, ActionPolicy)
 * @See BindingFeature#createBindingModule(EndpointActionReference, ActionPolicy, ActionPolicy)
 * @see TransportFeature#createTransportModule(EndpointActionReference, ActionPolicy, ActionPolicy)
 */
public final class ActionPolicy {

    private boolean isPrivate;
    private ServicePolicy servicePolicy;
    private EndpointPolicy endpointPolicy;
    private OperationPolicy operationPolicy;
    private MessagePolicy messagePolicy;

    /**
     * Constructs an ActionPolicy object
     * 
     * @param boolean
     *        isPrivate
     * 
     * @param servicePolicy
     *        effective policy with subject SERVICE or null if none defined.
     * 
     * @param endpointPolicy
     *        effective policy with subject ENDPOINT or null if none defined.
     * 
     * @param operationPolicy
     *        effective policy with subject OPERATION or null if none defined.
     * 
     * @param messagePolicy
     *        effective policy with subject MESSAGE or null if none defined.
     */
    public ActionPolicy(boolean isPrivate, ServicePolicy servicePolicy, EndpointPolicy endpointPolicy,
            OperationPolicy operationPolicy, MessagePolicy messagePolicy) {
        this.isPrivate = isPrivate;
        this.servicePolicy = servicePolicy;
        this.endpointPolicy = endpointPolicy;
        this.operationPolicy = operationPolicy;
        this.messagePolicy = messagePolicy;
    }

    /**
     * If true, the contained expressions were sourced from 'private' element policy, i.e.
     * attached to the following components:
     * <p>
     * {@link Engine#getEffectivePolicy()}
     * <p>
     * {@link ActorReference#getEffectivePolicy()}
     * <p>
     * {@link EndpointReference#getEffectivePolicy()}
     * <p>
     * {@link EndpointOperationReference#getEffectivePolicy()}
     * <p>
     * {@link EndpointActionReference#getEffectivePolicy()}
     * <p>
     * If false, the contained expressions were sourced from 'public' element policy, i.e.
     * attached to the following components:
     * <p>
     * {@link Service#getEffectivePolicy()}
     * <p>
     * {@link Endpoint#getEffectivePolicy()}
     * <p>
     * {@link EndpointOperation#getEffectivePolicy()}
     * <p>
     * {@link EndpointAction#getEffectivePolicy()}
     * <p>
     * 
     * @return true if private, false if public
     */
    public boolean isPrivate() {
        return isPrivate;
    }

    /**
     * Indicates whether or not container is empty.
     * 
     * @return true if at least one policy subject is defined, false if not.
     */
    public boolean isEmpty() {
        if (messagePolicy != null) {
            return false;
        }
        if (operationPolicy != null) {
            return false;
        }
        if (endpointPolicy != null) {
            return false;
        }
        if (servicePolicy != null) {
            return false;
        }
        return true;
    }

    /**
     * Returns effective policy with subject SERVICE.
     * 
     * @return ServicePolicy or <code>null</code> if no policy defined
     */
    public ServicePolicy getServicePolicy() {
        return servicePolicy;
    }

    /**
     * Returns effective policy with subject ENDPOINT.
     * 
     * @return EndpointPolicy or <code>null</code> if no policy defined
     */
    public EndpointPolicy getEndpointPolicy() {
        return endpointPolicy;
    }

    /**
     * Returns effective policy with subject OPERATION.
     * 
     * @return OperationPolicy or <code>null</code> if no policy defined
     */
    public OperationPolicy getOperationPolicy() {
        return operationPolicy;
    }

    /**
     * Returns effective policy with subject MESSAGE.
     * 
     * @return MessagePolicy or <code>null</code> if no policy defined
     */
    public MessagePolicy getMessagePolicy() {
        return messagePolicy;
    }

    /**
     * Convenience method which filters all contained effective expressions by invoking
     * {@link EffectivePolicy#filter(Set)} on each and returns result as a new
     * <code>ActionPolicy</code> instance.
     * 
     * @param policyVocabularyNamespaces
     *        filters all namespaces other than the contained 'understood' vocabularies
     * @return ActionPolicy
     */
    public ActionPolicy filter(Set<String> policyVocabularyNamespaces) {
        EffectivePolicy mp = messagePolicy == null ? null : messagePolicy.filter(policyVocabularyNamespaces);
        EffectivePolicy op = operationPolicy == null ? null : operationPolicy.filter(policyVocabularyNamespaces);
        EffectivePolicy ep = endpointPolicy == null ? null : endpointPolicy.filter(policyVocabularyNamespaces);
        EffectivePolicy sp = servicePolicy == null ? null : servicePolicy.filter(policyVocabularyNamespaces);
        return new ActionPolicy(isPrivate, (ServicePolicy)sp, (EndpointPolicy)ep, (OperationPolicy)op,
                (MessagePolicy)mp);
    }

}