/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.policy.rt;

import java.util.Set;


/**
 * Element policy may be attached to different components within the component hierarchy, each
 * of which, according to the ws-policy attachment spec, is called a <code>policy scope</code>.
 * This class merges expressions from all <code>policy scopes</code> associated with a
 * particular <code>PolicySubject</code> into a single 'effective' expression.
 */
public abstract class EffectivePolicy {

    protected PolicyExpression expression;

    static final String LINE_BREAK = System.getProperty("line.separator");

    public EffectivePolicy(PolicyExpression expression) {
        if (expression == null) {
            throw new IllegalArgumentException("expression arg required");
        }
        this.expression = expression;
    }

    /**
     * An enumeration of policy subjects. The subject to which a policy expression applies is
     * implied by the component to which it is attached.
     */
    public static enum PolicySubject {
        SERVICE, ENDPOINT, OPERATION, MESSAGE
    }

    /**
     * Identifies the policy subject, i.e. the subject constrained by assertions defined within
     * merged expression. All assertions defined within merged expression SHOULD describe
     * behaviors which directly affect the implied subject.
     * 
     * @return PolicySubject an enumerated value
     */
    public abstract PolicySubject getPolicySubject();

    /**
     * Returns merged policy expression for subject to which this effective policy applies.
     * Note that the returned expression is in partial normal form, i.e. nested policy
     * expressions, if any, are NOT normalized.
     * <p>
     * Note that policy expressions attached to subject MUST all be of same type, i.e. as
     * indicated by root element namespace, e.g. 'http://www.w3.org/ns/ws-policy'.
     * 
     * @return PolicyExpression effective expression
     */
    public PolicyExpression getExpression() {
        return expression;
    }

    /**
     * Convenience method which filters contained expression by invoking
     * {@link PolicyExpression#filter(Set)} and returning result as a new
     * <code>EffectivePolicy</code> instance.
     * 
     * @param policyVocabularyNamespaces
     *        filters all namespaces other than the contained 'understood' vocabularies
     * @return EffectivePolicy or <code>null</code> if result fails to contain at least one
     *         assertion with a matching namespace
     */
    public abstract EffectivePolicy filter(Set<String> policyVocabularyNamespaces);

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(LINE_BREAK);
        sb.append("policySubject: " + getPolicySubject());
        sb.append(LINE_BREAK);
        sb.append(expression.toString());
        return sb.toString();
    }

}