/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.service;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.BindingFault;
import org.bluestemsoftware.specification.eoa.component.binding.BindingFaultReference;
import org.bluestemsoftware.specification.eoa.component.binding.BindingMessageReference;
import org.bluestemsoftware.specification.eoa.component.binding.BindingOperation;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFault;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.policy.rt.MessagePolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyExpression;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;

/**
 * An invokable abstract action, i.e. an InterfaceAction which is bound and addressed.
 */
public final class EndpointAction extends NestedComponent<Service, EndpointOperation> {

    private String action;
    private InterfaceAction interfaceAction;
    private MessagePolicy effectivePolicy;
    private boolean effectivePolicyDerived;

    EndpointAction(String action) {
        if (action == null) {
            throw new IllegalArgumentException("action null");
        } else {
            this.action = action;
        }
    }

    /**
     * Retrieves value which uniquely identifies abstract action bound by <code>Binding</code>
     * employed by parent Endpoint
     * 
     * @return action
     * 
     * @see InterfaceAction#getAction()
     */
    public String getAction() {
        return action;
    }

    /**
     * Gets <code>InterfaceAction</code> bound by <code>Binding</code> employed by parent
     * Endpoint
     * 
     * @return
     */
    public InterfaceAction getInterfaceAction() {
        return interfaceAction;
    }

    /**
     * Returns 'effective' policy with policy subject MESSAGE from the perspective of parent
     * endpoint. The methodology used to calculate effective policy depends upon whether or not
     * referenced abstract action is associated with a normal message or with a fault message.
     * <p>
     * If action is associated with a normal message, effective Policy is the merged result of
     * element Policy attached to corresponding {@link BindingMessageReference}, to
     * corresponding {@link InterfaceMessageReference} and to {@link InterfaceMessage}
     * referenced by <code>InterfaceMessageReference</code>.
     * <p>
     * If action is associated with a fault message, effective Policy is the merged result of
     * element Policy attached to corresponding {@link BindingFaultReference}, to
     * {@link BindingFault} referenced by <code>BindingFaultReference</code>, to
     * {@link InterfaceFaultReference} bound by <code>BindingFaultReference</code>, to
     * {@link InterfaceFault} referenced by <code>InterfaceFaultReference</code> and to
     * {@link InterfaceMessage} referenced by <code>InterfaceFault</code>.
     * 
     * @return EffectivePolicy effective policy with policy subject MESSAGE or
     *         <code>null</code> if no policy defined.
     */
    public synchronized MessagePolicy getEffectivePolicy() throws PolicyException {
        if (effectivePolicyDerived) {
            return effectivePolicy;
        }
        effectivePolicy = deriveEffectivePolicy();
        effectivePolicyDerived = true;
        return effectivePolicy;
    }

    private MessagePolicy deriveEffectivePolicy() throws PolicyException {

        PolicyExpression mergedExpression = null;

        Endpoint endpoint = parent.getParent();
        Binding binding = endpoint.getBindingReference().getReferencedComponent();
        BindingOperation bindingOperation = binding.getOperation(parent.getName());
        Interface intrface = getRootComponent().getInterfaceReference().getReferencedComponent();
        InterfaceAction interfaceAction = intrface.getInterfaceAction(action);

        if (interfaceAction instanceof InterfaceMessageReference) {
            InterfaceMessageReference imr = (InterfaceMessageReference)interfaceAction;
            if (imr.getElementPolicy() != null) {
                PolicyExpression imrlp = imr.getElementPolicy().getMergedExpression();
                mergedExpression = (mergedExpression == null) ? imrlp : mergedExpression.mergeExpression(imrlp);
            }
            InterfaceMessage interfaceMessage = imr.getReferencedComponent();
            if (interfaceMessage != null && interfaceMessage.getElementPolicy() != null) {
                PolicyExpression imlp = interfaceMessage.getElementPolicy().getMergedExpression();
                mergedExpression = (mergedExpression == null) ? imlp : mergedExpression.mergeExpression(imlp);
            }
            if (bindingOperation != null) {
                BindingMessageReference bmr = null;
                if (interfaceAction.getDirection() == Direction.IN) {
                    bmr = bindingOperation.getInputMessageReference();
                } else {
                    bmr = bindingOperation.getOutputMessageReference();
                }
                if (bmr != null && bmr.getElementPolicy() != null) {
                    PolicyExpression bmrlp = bmr.getElementPolicy().getMergedExpression();
                    mergedExpression = (mergedExpression == null) ? bmrlp : mergedExpression
                            .mergeExpression(bmrlp);
                }
            }
        } else {
            InterfaceFaultReference ifr = (InterfaceFaultReference)interfaceAction;
            if (ifr.getElementPolicy() != null) {
                PolicyExpression ifrlp = ifr.getElementPolicy().getMergedExpression();
                mergedExpression = (mergedExpression == null) ? ifrlp : mergedExpression.mergeExpression(ifrlp);
            }
            InterfaceFault interfaceFault = ifr.getFault();
            if (interfaceFault.getElementPolicy() != null) {
                PolicyExpression iflp = interfaceFault.getElementPolicy().getMergedExpression();
                mergedExpression = (mergedExpression == null) ? iflp : mergedExpression.mergeExpression(iflp);
            }
            InterfaceMessage interfaceMessage = interfaceFault.getReferencedComponent();
            if (interfaceMessage != null && interfaceMessage.getElementPolicy() != null) {
                PolicyExpression imlp = interfaceMessage.getElementPolicy().getMergedExpression();
                mergedExpression = (mergedExpression == null) ? imlp : mergedExpression.mergeExpression(imlp);
            }
            if (bindingOperation != null) {
                BindingFaultReference bfr = bindingOperation.getFaultReference(interfaceFault.getName());
                if (bfr != null) {
                    if (bfr.getElementPolicy() != null) {
                        PolicyExpression bfrlp = bfr.getElementPolicy().getMergedExpression();
                        mergedExpression = (mergedExpression == null) ? bfrlp : mergedExpression
                                .mergeExpression(bfrlp);
                    }
                    BindingFault bindingFault = bfr.getFault();
                    if (bindingFault.getElementPolicy() != null) {
                        PolicyExpression bflp = bindingFault.getElementPolicy().getMergedExpression();
                        mergedExpression = (mergedExpression == null) ? bflp : mergedExpression
                                .mergeExpression(bflp);
                    }
                }
            }
        }

        if (mergedExpression != null) {
            return new MessagePolicy(mergedExpression.normalizeExpression());
        } else {
            return null;
        }

    }

    final void establishReference() throws DeploymentException {
        if (interfaceAction != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        Interface intrface = getRootComponent().getInterfaceReference().getReferencedComponent();
        interfaceAction = intrface.getInterfaceAction(action);
        if (interfaceAction == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. InterfaceAction '"
                    + action
                    + "', not defined on Interface '"
                    + intrface.getName()
                    + "'.");
        }
    }

    final void validate() throws DeploymentException {
        // TODO: implement
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(action);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "endpointAction",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(EndpointOperation parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("action: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(action);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(action);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof EndpointAction)) {
            return false;
        }

        if ((action == ((EndpointAction)obj).action || (action != null && action
                .equals(((EndpointAction)obj).action)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}