/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.service;

import java.util.Collection;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.BindingOperation;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation;
import org.bluestemsoftware.specification.eoa.component.policy.rt.OperationPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyExpression;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;

public final class EndpointOperation extends NestedComponent<Service, Endpoint> {

    private QName name;
    private InterfaceOperation interfaceOperation;
    private Map<String, EndpointAction> endpointActions;
    private OperationPolicy effectivePolicy;
    private boolean effectivePolicyDerived;

    EndpointOperation(QName name, Map<String, EndpointAction> endpointActions) {
        this.name = name;
        this.endpointActions = endpointActions;
        for (EndpointAction endpointAction : endpointActions.values()) {
            endpointAction.setParent(this);
        }
    }

    /**
     * Identifies name of referenced <code>InterfaceOperation</code> component.
     * 
     * @return QName name of referenced <code>InterfaceOperation</code>.
     */
    public QName getName() {
        return name;
    }

    /**
     * Retrieves an action defined on this endpoint.
     * 
     * @param action
     *        URI
     * @return EndpointAction
     * 
     * @see InterfaceAction#getAction()
     */
    public EndpointAction getEndpointAction(String action) {
        return endpointActions.get(action);
    }

    /**
     * Retrieves all actions defined on this endpoint.
     * 
     * @return Collection of EndpointAction components or an empty collection if none defined
     * 
     * @see InterfaceAction#getAction()
     */
    public Collection<EndpointAction> getEndpointActions() {
        return endpointActions.values();
    }

    /**
     * Returns 'effective' policy, i.e. returns the merged result of element Policy attached to
     * corresponding BindingOperation on Binding utilized by this Endpoint and element policy
     * attached to corresponding InterfaceOperation on Interface implemented by parent Service
     * component.
     * 
     * @return EffectivePolicy effective policy with policy subject OPERATION or
     *         <code>null</code> if no policy defined.
     */
    public synchronized OperationPolicy getEffectivePolicy() throws PolicyException {
        if (effectivePolicyDerived) {
            return effectivePolicy;
        }
        effectivePolicy = deriveEffectivePolicy();
        effectivePolicyDerived = true;
        return effectivePolicy;
    }
    
    /**
     * Gets referenced <code>InterfaceOperation</code> component.
     * 
     * @return InterfaceOperation referenced <code>InterfaceOperation</code>.
     */
     public InterfaceOperation getInterfaceOperation() {
         return interfaceOperation;
     }
    
    void establishReference() throws DeploymentException {
        if (interfaceOperation != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        Interface intrface = getRootComponent().getInterfaceReference().getReferencedComponent();
        interfaceOperation = intrface.getOperation(name);
        if (interfaceOperation == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. InterfaceOperation '"
                    + name
                    + "', not defined on Interface '"
                    + intrface.getName()
                    + "'.");
        }
        for (EndpointAction endpointAction : endpointActions.values()) {
            endpointAction.establishReference();
        }
    }

    private OperationPolicy deriveEffectivePolicy() throws PolicyException {

        PolicyExpression mergedExpression = null;

        Binding binding = parent.getBindingReference().getReferencedComponent();
        BindingOperation bindingOperation = binding.getOperation(name);

        if (bindingOperation != null && bindingOperation.getElementPolicy() != null) {
            PolicyExpression bolp = bindingOperation.getElementPolicy().getMergedExpression();
            mergedExpression = (mergedExpression == null) ? bolp : mergedExpression.mergeExpression(bolp);
        }

        Interface intrface = getRootComponent().getInterfaceReference().getReferencedComponent();
        InterfaceOperation interfaceOperation = intrface.getOperation(name);

        if (interfaceOperation.getElementPolicy() != null) {
            PolicyExpression iolp = interfaceOperation.getElementPolicy().getMergedExpression();
            mergedExpression = (mergedExpression == null) ? iolp : mergedExpression.mergeExpression(iolp);
        }

        if (mergedExpression != null) {
            return new OperationPolicy(mergedExpression.normalizeExpression());
        } else {
            return null;
        }

    }
    
    void validate() throws DeploymentException {
        for(EndpointAction endpointAction : endpointActions.values()) {
            endpointAction.validate();
        }
        // TODO: validate myself
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(name.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + name.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "endpointOperation", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(Endpoint parent) {
        this.parent = parent;
    }
    
    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("endpointActions: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(endpointActions);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(name) + hashCode(endpointActions);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof EndpointOperation)) {
            return false;
        }

        if (
                (name == ((EndpointOperation)obj).name || (name != null && name.equals(
                ((EndpointOperation)obj).name)))
                &&
                (endpointActions == ((EndpointOperation)obj).endpointActions || (endpointActions != null && endpointActions.equals(
                        ((EndpointOperation)obj).endpointActions)))         
        ) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }
}