/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.service;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ServicePolicy;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;

/**
 * Models a Service component. A Service component describes a set of invokable endpoints each
 * of which identifies a protocol specific location.
 */
public final class Service extends RootComponent {

    private boolean referencesEstablished;
    private InterfaceReference interfaceReference;
    private ElementPolicy elementPolicy;
    private ServicePolicy effectivePolicy;
    private Map<String, Endpoint> endpoints;
    private boolean effectivePolicyDerived;

    public Service(ComponentName name, InterfaceReference interfaceReference, ElementPolicy elementPolicy,
            Map<String, Endpoint> endpoints) {

        if (name == null) {
            throw new IllegalArgumentException("serviceName null");
        } else {
            this.name = name;
        }
        fragmentIdentifier = new FragmentIdentifier(name.getNamespaceURI(), Scheme.EOA, "service", name
                .getLocalPart());
        if (interfaceReference == null) {
            throw new IllegalArgumentException("interfaceName null");
        } else {
            this.interfaceReference = interfaceReference;
            interfaceReference.setParent(this);
        }
        if (endpoints == null) {
            throw new IllegalArgumentException("endpoints null");
        } else {
            this.endpoints = Collections.unmodifiableMap(endpoints);
            for (Endpoint endpoint : endpoints.values()) {
                endpoint.setParent(this);
            }
        }
        this.elementPolicy = elementPolicy;

    }

    @Override
    public final ComponentType getComponentType() {
        return ComponentType.SERVICE;
    }

    /**
     * Returns reference to <code>Interface</code> implemented by this service.
     * 
     * @return InterfaceReference
     */
    public InterfaceReference getInterfaceReference() {
        return interfaceReference;
    }

    /**
     * Metadata describing a specific Endpoint on this service.
     * <p>
     * 
     * @return Endpoint
     */
    public final Endpoint getEndpoint(String endpointName) {
        return endpoints.get(endpointName);
    }

    /**
     * Metadata describing all endpoints defined on this service.
     * <p>
     * 
     * @return Collection of endpointInfo objects or an empty Collection if none defined.
     */
    public Collection<Endpoint> getEndpoints() {
        return endpoints.values();
    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * 
     * @return ElementPolicy or <code>null</code> if no policy attached.
     */
    public ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /**
     * Returns the 'effective' policy for this Component, which in this case represents element
     * policy attached to this component.
     * 
     * @return EffectivePolicy effective policy with policy subject SERVICE or
     *         <code>null</code> if no policy defined.
     */
    public synchronized ServicePolicy getEffectivePolicy() throws PolicyException {
        if (effectivePolicyDerived) {
            return effectivePolicy;
        }
        effectivePolicy = deriveEffectivePolicy();
        effectivePolicyDerived = true;
        return effectivePolicy;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#validate()
     */
    @Override
    protected void validate() throws DeploymentException {
        
        for (Endpoint endpoint : endpoints.values()) {
            endpoint.validate();
        }    
        
        // TODO: validate myself
        
        // TODO: a service should not be hosted by more than one engine. not sure
        // how to validate this. currently an error will be thrown by transport
        // if an endpoint address is mapped to more than one transport, e.g. a
        // servlet path, which catches this error
        
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#establishReferences()
     */
    @Override
    protected synchronized void establishReferences() throws DeploymentException {
        if (referencesEstablished) {
            return;
        }
        interfaceReference.establishReference();
        for (Endpoint endpoint : endpoints.values()) {
            endpoint.establishReference();
        }
        try {
            deriveEffectivePolicy();
        } catch (PolicyException pe) {
            throw new DeploymentException(pe);
        }
        referencesEstablished = true;
    }

    /*
     * Aggregates element policy expressions by subject. Merge operation forces resolution of
     * policy refeferences which MAY occur within policy expressions, i.e. references to policy
     * components and or remotely defined expressions
     */
    private ServicePolicy deriveEffectivePolicy() throws PolicyException {
        if (elementPolicy != null) {
            return new ServicePolicy(elementPolicy.getMergedExpression().normalizeExpression());
        }
        return null;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("interfaceReference: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(interfaceReference);
        sb.append(System.getProperty("line.separator"));
        sb.append("endpoints: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(endpoints);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(name) + hashCode(interfaceReference) + hashCode(endpoints);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Service)) {
            return false;
        }

        if ((name == ((Service)obj).name || (name != null && name.equals(((Service)obj).name)))
                && (interfaceReference == ((Service)obj).interfaceReference || (interfaceReference != null && interfaceReference
                        .equals(((Service)obj).interfaceReference)))
                && (endpoints == ((Service)obj).endpoints || (endpoints != null && endpoints
                        .equals(((Service)obj).endpoints)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
