/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.types;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;

/**
 * Models a schema component, i.e. an element declaration, a complex type definition or a
 * simple type definition.
 */

public abstract class SchemaComponent extends NestedComponent<Schema, Schema> {

    protected String name;
    protected QName qName;
    protected Map<QName, SchemaFragment> fragments;
    
    /**
     * An enumeration indicating schema component types modeled by this specification. The
     * types are defined as follows:
     * <p>
     * ELEMENT_DECLARATION - SchemaComponent models a global element declaration
     * <p>
     * TYPE_DEFINITION - SchemaComponent models a named complex type definition or a simple
     * type definition.
     */
    public static enum Type {
        ELEMENT_DECLARATION, TYPE_DEFINITION
    }

    public SchemaComponent(String name, Map<QName, SchemaFragment> fragments) {
        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        if (fragments == null) {
            throw new IllegalArgumentException("fragments null");
        } else {
            this.fragments = Collections.unmodifiableMap(fragments);
        }
        for (SchemaFragment fragment : fragments.values()) {
            fragment.setParent(this);
        }
    }

    /**
     * Gets <code>SchemaComponent</code> name, which MUST be unique within context of parent
     * <code>Schema</code> for indicated type.
     * 
     * @return name
     */
    public String getName() {
        return name;
    }
    
    /**
     * Convenience method which retrieves qualified name of <code>SchemaComponent</code>.
     * @return
     */
    public QName getQName() {
        return qName;
    }

    /**
     * Retrieves value which indicates schema component type.
     * 
     * @return Enumerated value
     */
    public abstract Type getType();

    /**
     * Gets <code>Fragment</code> which statically defines an embedded instance of indicated
     * <code>Property</code> which may be extracted from a runtime instance of
     * <code>SchemaComponent</code>, e.g. an element information item.
     * 
     * @param propertyName
     * @return Fragment or null if fragment is undefined
     */
    public SchemaFragment getFragment(QName propertyName) {
        return fragments.get(propertyName);
    }

    /**
     * Gets all <code>Fragment</code> components defined on this component.
     * 
     * @return Collection of Fragments or an empty collection if none defined.
     */
    public Collection<SchemaFragment> getFragments() {
        return fragments.values();
    }

    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(getType());
            pointerPart.append("/");
            pointerPart.append(name);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "schemaComponent",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }
    
    void validate() throws DeploymentException {
        for (SchemaFragment fragment : fragments.values()) {
            fragment.validate();
        }    
    }

    void setParent(Schema parent) {
        this.parent = parent;
        this.qName = new QName(parent.getName().getNamespaceURI(), name);
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("type:");
        sb.append(getType());
        sb.append(System.getProperty("line.separator"));
        sb.append("fragments:");
        sb.append(fragments);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(name) + hashCode(getType())+ hashCode(fragments);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof SchemaComponent)) {
            return false;
        }

        if ((name == ((SchemaComponent)obj).name || (name != null && name.equals(((SchemaComponent)obj).name)))
                && (getType() == ((SchemaComponent)obj).getType() || (getType() != null && getType().equals(
                        ((SchemaComponent)obj).getType())))
                        && (fragments == ((SchemaComponent)obj).fragments || (fragments != null && fragments.equals(
                                ((SchemaComponent)obj).fragments)))        ) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}