/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.types.rt;

import java.util.Map;

import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation.Style;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT.SingletonRT;
import org.bluestemsoftware.specification.eoa.component.types.Schema;
import org.bluestemsoftware.specification.eoa.ext.RootExtension;
import org.bluestemsoftware.specification.eoa.ext.schema.SchemaFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.ext.LexicalHandler;

/**
 * Models a schema document. The language used to model <code>SchemaDocument</code> is an
 * extension to the core specification, e.g. W3C XML Schema, relax-ng, schematron, etc ...
 * <p>
 * Note that it is implementation's responsibility to switch thread context classloader as
 * required to/from {@link SchemaFactory} deployment classloader.
 * 
 * @see SchemaFactory
 */
public interface SchemaDocument extends RootExtension, SingletonRT<Schema>, Document {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends RootExtension.Provider {
    }

    /**
     * Returns 'target namespace', i.e. namespace within which all contained element decls and
     * type definitions are defined.
     * <p>
     * Note that although some schema languages, e.g. relax-ng, may support schema documents
     * with multiple namespaces, this specification requires that a document define exactly one
     * target namespace.
     * 
     * @see http://www.w3.org/TR/wsdl20-altschemalangs/
     * 
     * @return String URI
     */
    public String getTargetNamespace();

    /**
     * Validates an element instance against its definition, i.e. against an element
     * declaration declared within this <code>SchemaDocument</code>.
     * 
     * @param content
     * @throws SchemaValidationException
     *         if content is not schema valid.
     * @throws SchemaException
     *         if an error occurs performing validation.
     */
    public void validate(Element content) throws SchemaException, SchemaValidationException;

    /**
     * Validates an element instance against its definition, i.e. against an element
     * declaration declared within this <code>SchemaDocument</code>.
     * 
     * @param content
     * @param contentHandler
     * @param lexicalHandler
     * @param errorHandler
     * @throws SchemaException
     *         if an error occurs performing validation.
     */
    public void validate(Element content, ContentHandler contentHandler, LexicalHandler lexicalHandler, ErrorHandler errorHandler) throws SchemaException;

    /**
     * Clones this <code>SchemaDocument</code> instance.
     * 
     * @param documentURI
     *        Identifies location from which document is being dereferenced. When cloning value
     *        of location attribute on schema imports and includes, implementations MUST use
     *        documentURI as base against which a relative URI, which is a URI fragment whose
     *        value is fragmentIdentifier of referenced schema, is resolved to create location
     *        URL.
     * 
     * @return Document
     */
    public Document clone(String documentURI) throws SchemaException;

    /**
     * Validates that schema component associated with normal, i.e. non-fault, action conforms
     * to style(s) defined on parent operation/interface and returns object(s) which models
     * each indicated style.
     * <p>
     * Note that a compliant implementation MUST process all 'predefined styles' defined by
     * WSDL 2.0 adjuncts specification.
     * 
     * @param interfaceAction
     *        abstract action for which style(s) are to be modeled.
     * @return Collection of <code>Style</code> objects each of which is mapped to IRI which
     *         uniquely identifies style, or an empty map if none defined/applicable.
     * @throws SchemaException
     *         If style(s) not understood or if associated schema component fails to conform to
     *         style or if an error occurs generating model(s).
     *         
     * @see http://www.w3.org/TR/wsdl20-adjuncts/#styles        
     */
    public Map<String, Style> modelStyles(InterfaceMessageReference interfaceMessageReference) throws SchemaException;

}
