/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.wsa.EndpointReference;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * Defines abstract ws-addressing message properties. Note that <code>WSAMessageModule</code>
 * definition is built-in to this specification and as such, an instance MUST be defined within
 * each instance of {@link MessageModules}.
 */
public final class WSAMessageModule extends MessageModule {

    public static final String TYPE = WSAMessageModule.class.getName();

    private String action;
    private String namespace;
    private String to;
    private String messageID;
    private EndpointReference replyTo;
    private EndpointReference faultTo;
    private EndpointReference from;
    private String relatesTo;
    private Set<Element> referenceParameters;

    /**
     * Constructs <code>MessageModule</code> instance. Generates a unique messageID.
     * 
     * @param action
     *        defines action
     */
    public WSAMessageModule(String action) {
        this(action, "urn:uuid:" + UUID.randomUUID().toString());
    }

    /**
     * Constructs <code>MessageModule</code> instance with provided messageID.
     * 
     * @param action
     *        defines action
     * @param messageID
     *        A globally unique message id.
     */
    public WSAMessageModule(String action, String messageID) {
        if (action == null || action.equals("")) {
            throw new IllegalArgumentException("action invalid");
        }
        if (messageID == null || messageID.equals("")) {
            throw new IllegalArgumentException("messageID invalid");
        }
        this.action = action;
        this.messageID = messageID;
    }

    /**
     * Identifes the semantics implied by the associated message, which SHOULD match an action
     * defined on <code>Interface</code> implemented by target <code>Service</code>.
     * 
     * @return action URI which MUST NOT be <code>null</code>
     * 
     * @see InterfaceAction
     */
    public String getAction() {
        return action;
    }

    /**
     * If associated message represents a request, identifies intended receiver for fault
     * responses.
     * 
     * @return
     */
    public EndpointReference getFaultTo() {
        return faultTo;
    }

    /**
     * A reference to endpoint from which the associated message originated.
     * 
     * @return
     */
    public EndpointReference getFrom() {
        return from;
    }

    /**
     * Gets ID which Uniquely identifies the associated message.
     * @return
     */
    public String getMessageID() {
        return messageID;
    }

    /**
     * If associated message represents a response, returns message id of related request
     * message. Note that the MEPs supported by this specification preclude there being more
     * than one relatesTo value.
     * 
     * @return RelatesTo or <code>null</code> if undefined.
     */
    public String getRelatesTo() {
        return relatesTo;
    }

    /**
     * If associated message represents a request, identifies intended receiver for response.
     * 
     * @return
     */
    public EndpointReference getReplyTo() {
        return replyTo;
    }

    /**
     * Identifies message destination, i.e. the address of the intended receiver of associated
     * message.
     * 
     * @return an absolute IRI
     */
    public String getTo() {
        return to;
    }

    /**
     * Gets reference parameters
     * @return reference parameters, or an empty set if none defined.
     */
    @SuppressWarnings("unchecked")
    public Set<Element> getReferenceParameters() {
        if (referenceParameters == null) {
            return Collections.EMPTY_SET;
        }
        return referenceParameters;
    }

    public void setFaultTo(EndpointReference faultTo) {
        this.faultTo = faultTo;
    }

    public void setFrom(EndpointReference from) {
        this.from = from;
    }

    public void setNamespace(String namespace) {
        this.namespace = namespace;
    }

    public void setReferenceParameters(Set<Element> referenceParameters) {
        this.referenceParameters = referenceParameters;
    }

    public void setRelatesTo(String relatesTo) {
        this.relatesTo = relatesTo;
    }

    public void setReplyTo(EndpointReference replyTo) {
        this.replyTo = replyTo;
    }

    public void setTo(String to) {
        this.to = to;
    }

    public void setAction(String action) {
        this.action = action;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule#getNamespace()
     */
    @Override
    public String getNamespace() {
        return namespace;
    }

    /**
     * Converts an <code>Element</code> into a type specific instanceof
     * <code>WSAMessageModule</code>.
     * 
     * @param element
     *        element instance to convert
     * @return
     */
    public static WSAMessageModule valueOf(Element element) {

        if (element == null || !element.getNamespaceURI().equals(TYPE)) {
            throw new IllegalArgumentException("Expected element of type " + new QName(TYPE, "module"));
        }

        String namespace = element.getAttribute("namespace");
        String action = getText(getChildElement(element, "action"));
        String messageID = getText(getChildElement(element, "messageID"));

        WSAMessageModule wsamm = new WSAMessageModule(action, messageID);
        wsamm.setNamespace(namespace);

        Element child = null;
        child = getChildElement(element, "to");
        if (child != null) {
            wsamm.setTo(getText(child));
        }
        child = getChildElement(element, "referenceParameters");
        if (child != null) {
            wsamm.setReferenceParameters(getReferenceParameters(child));
        }
        child = getChildElement(element, "replyTo");
        if (child != null) {
            String address = child.getAttribute("address");
            Element parmsElement = getChildElement(child, "referenceParameters");
            Set<Element> referenceParameters = null;
            if (parmsElement != null) {
                referenceParameters = getReferenceParameters(parmsElement);
            }
            wsamm.setReplyTo(new EndpointReference(address, referenceParameters));
        }
        child = getChildElement(element, "faultTo");
        if (child != null) {
            String address = child.getAttribute("address");
            Element parmsElement = getChildElement(child, "referenceParameters");
            Set<Element> referenceParameters = null;
            if (parmsElement != null) {
                referenceParameters = getReferenceParameters(parmsElement);
            }
            wsamm.setFaultTo(new EndpointReference(address, referenceParameters));
        }
        child = getChildElement(element, "from");
        if (child != null) {
            String address = child.getAttribute("address");
            Element parmsElement = getChildElement(child, "referenceParameters");
            Set<Element> referenceParameters = null;
            if (parmsElement != null) {
                referenceParameters = getReferenceParameters(parmsElement);
            }
            wsamm.setFrom(new EndpointReference(address, referenceParameters));
        }
        child = getChildElement(element, "relatesTo");
        if (child != null) {
            String relatesToMessageID = child.getAttribute("messageID");
            wsamm.setRelatesTo(relatesToMessageID);
        }
        return wsamm;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule#toElement(org.w3c.dom.Document)
     */
    @Override
    public Element toElement(Document ownerDocument) {
        Element wrapper = ownerDocument.createElementNS(TYPE, "wsa:module");
        wrapper.setAttribute("namespace", namespace);
        if (to != null) {
            Element child = ownerDocument.createElementNS(TYPE, "wsa:to");
            child.appendChild(ownerDocument.createTextNode(to));
            wrapper.appendChild(child);
        }
        if (action != null) {
            Element child = ownerDocument.createElementNS(TYPE, "wsa:action");
            child.appendChild(ownerDocument.createTextNode(action));
            wrapper.appendChild(child);
        }
        if (messageID != null) {
            Element child = ownerDocument.createElementNS(TYPE, "wsa:messageID");
            child.appendChild(ownerDocument.createTextNode(messageID));
            wrapper.appendChild(child);
        }
        if (referenceParameters != null && referenceParameters.size() > 0) {
            Element child = ownerDocument.createElementNS(TYPE, "wsa:referenceParameters");
            wrapper.appendChild(child);
            for (Element parmElement : referenceParameters) {
                child.appendChild(ownerDocument.importNode(parmElement, true));
            }
        }
        if (replyTo != null) {
            Element child = ownerDocument.createElementNS(TYPE, "wsa:replyTo");
            child.setAttribute("address", replyTo.getAddress());
            wrapper.appendChild(child);
            Element parmsElement = null;
            for (Element parmElement : replyTo.getReferenceParameters()) {
                if (parmsElement == null) {
                    parmsElement = ownerDocument.createElementNS(TYPE, "referenceParameters");
                    child.appendChild(parmsElement);
                }
                parmsElement.appendChild(ownerDocument.importNode(parmElement, true));
            }
        }
        if (faultTo != null) {
            Element child = ownerDocument.createElementNS(TYPE, "wsa:faultTo");
            child.setAttribute("address", faultTo.getAddress());
            wrapper.appendChild(child);
            Element parmsElement = null;
            for (Element parmElement : faultTo.getReferenceParameters()) {
                if (parmsElement == null) {
                    parmsElement = ownerDocument.createElementNS(TYPE, "referenceParameters");
                    child.appendChild(parmsElement);
                }
                parmsElement.appendChild(ownerDocument.importNode(parmElement, true));
            }
        }
        if (from != null) {
            Element child = ownerDocument.createElementNS(TYPE, "wsa:from");
            child.setAttribute("address", from.getAddress());
            wrapper.appendChild(child);
            Element parmsElement = null;
            for (Element parmElement : from.getReferenceParameters()) {
                if (parmsElement == null) {
                    parmsElement = ownerDocument.createElementNS(TYPE, "referenceParameters");
                    child.appendChild(parmsElement);
                }
                parmsElement.appendChild(ownerDocument.importNode(parmElement, true));
            }
        }
        if (relatesTo != null) {
            Element child = ownerDocument.createElementNS(TYPE, "wsa:relatesTo");
            child.setAttribute("messageID", relatesTo);
            wrapper.appendChild(child);
        }
        return wrapper;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule#getType()
     */
    @Override
    public String getType() {
        return TYPE;
    }

    private static Element getChildElement(Element parent, String childName) {
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (element.getLocalName().equals(childName)) {
                return element;
            }
        }
        return null;
    }

    private static Set<Element> getReferenceParameters(Element parent) {
        Set<Element> parameters = new HashSet<Element>();
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            parameters.add((Element)parent.removeChild(nodeList.item(i)));
        }
        return parameters;
    }

    private static String getText(Element element) {
        if (element == null || !element.hasChildNodes()) {
            return null;
        }
        try {
            Text textNode = (Text)element.getFirstChild();
            return textNode.getData();
        } catch (ClassCastException ce) {
        }
        return null;
    }

}
