/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.idl;

import javax.xml.transform.Source;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.ComponentReader;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactory;
import org.w3c.dom.Element;

/**
 * A factory which generates <code>IDLDocument</code> instances. Factory is specific to a
 * particular type of IDL, e.g. WSDL 1.1, WSDL 2.0, WADL, etc ...
 * <p>
 * Note that type ref URI retrieved via {@link #getExtensionType()} MUST match IDL namespace
 * URI defined by IDL specification which corresponds to implied version.
 */

public abstract class IDLFactory extends ExtensionFactory implements ComponentReader {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ExtensionFactory.Provider {
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExtensionFactory#getType()
     */
    public final Type getType() {
        return Type.IDL;
    }

    /**
     * Reads <code>Source</code> and returns a language specific <code>IDLDocument</code>
     * instance.
     * 
     * @param componentContext
     * @param source
     *        or <code>null</code> in which case a compliant factory MUST return an empty
     *        <code>IDLDocument</code>.
     *        <p>
     *        Note that if source is instanceof <code>SAXSource</code> which supplies an
     *        <code>XMLReader</code>, the reader MAY be ignored in preference to parser
     *        supplied by implementation.
     * @return
     * @throws IDLException
     *         if an error occurs reading document
     */
    public abstract IDLDocument readIDL(ComponentContext componentContext, Source source) throws IDLException;

    /**
     * Used by instances of <code>IDLFactory</code> to read a <code>Binding</code>
     * extension.
     */
    public interface BindingDefinitionReader {

        /**
         * Identifies 'binding style' inferred by concrete message format used by operations
         * defined on binding.
         * <p>
         * Note that basic profile requires that all operations defined on a WSDL 1.1 binding
         * utilize a consistent message style, while WSDL 2.0 operations MUST use only document
         * style messages.
         */
        public static enum BindingStyle {
            DOCUMENT_STYLE, RPC_STYLE
        }

        /**
         * Reads a <code>Binding</code> extension and corresponding runtime and adds
         * both to <code>ComponentDeployment</code>.
         * 
         * @param componentContext
         *        defines context of deployment from which definition is sourced.
         *        Implementations MUST use context to resolve entities referenced by
         *        components, if any.
         *        <p>
         *        Note that deployment MAY contain definitions for dependent components which
         *        MAY be required to parse Binding, e.g. a WSDL 1.1 Binding requires access to
         *        WSDL 1.1 interface bound by Binding.
         * 
         * @param targetNamespace
         *        of IDL document which contains binding definition
         * @param bindingElement
         *        wsdl version specific element instance
         * @param bindingProfile
         *        profile metadata which MAY be required by factory to generate Binding.
         * @return Binding which will be added to <code>ComponentDeployment</code> by caller.
         * @throws DeploymentException
         *         If an error occurs reading <code>Binding</code>, or if wsdl version not
         *         supported, i.e. as implied by bindingElement namespace.
         * 
         * @see ComponentDeployment#addComponent(RootComponent)
         * @see ComponentDeployment#addProvider(RootComponentRT)
         */
        public void readBinding(ComponentContext componentContext, String targetNamespace, Element bindingElement, BindingStyle bindingStyle) throws DeploymentException;

    }

    /**
     * Used by instances of <code>WSDLFactory</code> to write a WSDL <code>Binding</code>
     * extension.
     */
    public interface BindingWriter {

        /**
         * Writes a WSDL <code>Binding</code> extension.
         * 
         * @param wsdlDocument
         *        document to which definition should be written
         * @param binding
         * @throws IDLException
         *         if an error occurs writing component or if wsdl version not understood, i.e.
         *         as indicated by {@link IDLDocument#getExtensionType()}.
         */

        public void writeBinding(IDLDocument wsdlDocument, Binding binding) throws IDLException;

    }

}