/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.schema;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.ComponentReader;
import org.bluestemsoftware.specification.eoa.component.rt.ProviderReader;
import org.bluestemsoftware.specification.eoa.component.types.Schema;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaDocument;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaException;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactory;
import org.w3c.dom.Document;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * A schema factory used to create <code>SchemaDocument</code> components of a specific type,
 * i.e. language, e.g. xml schema, relax ng, ...
 */

public abstract class SchemaFactory extends ExtensionFactory implements ComponentReader, ProviderReader<SchemaDocument> {

    public static final String EOA_SCHEMA_EXT_NS = "http://bluestemsoftware.org/specification/eoa/1.0/component/schema/ext";
    public static final String EOA_SCHEMA_EXT_NS_PREFIX = "eoa";

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ExtensionFactory.Provider {
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExtensionFactory#getType()
     */
    public final Type getType() {
        return Type.SCHEMA;
    }

    /**
     * Parses <code>Source</code> into a <code>SchemaDocument</code>.
     * 
     * @param entityResolver
     *        which MUST be used to resolve external entities, if any.
     * @param source
     *        or <code>null</code> in which case a compliant factory MUST return a
     *        <code>Set</code> which contains a single empty <code>SchemaDocument</code>.
     *        <p>
     *        Note that if source is instanceof <code>SAXSource</code> which supplies an
     *        <code>XMLReader</code>, the reader MAY be ignored in preference to parser
     *        supplied by implementation.
     * @return A <code>Set</code> of schema documents which MUST include target document,
     *         i.e. the document manifested as <code>Source</code>, and all dereferenceable
     *         schema documents referenced by target document, if any.
     * 
     * @throws DeploymentException
     * 
     */
    public abstract Set<SchemaDocument> readSchemas(EntityResolver entityResolver, Source source) throws SchemaException;

    /**
     * Used by {@link ComponentDeployment} to build 'implied' <code>Schema</code> definition
     * using corresponding runtime provider definition.
     * <p>
     * Note that <code>Schema</code> is a special case, i.e. all other
     * <code>RootComponent</code> defintions MUST be sourced from an <code>IDLDocument</code>
     * and read via {@link ComponentReader} interface.
     * 
     * @param deployment
     * @param provider
     * @return
     * @throws DeploymentException
     */
    public abstract Schema buildSchema(ComponentDeployment deployment, SchemaDocument provider) throws DeploymentException;

    protected class DocumentParser implements ErrorHandler {

        private ArrayList<String> errorMessages = new ArrayList<String>();
        private DocumentBuilderFactory factory;

        public DocumentParser() throws Exception {
            factory = DocumentBuilderFactory.newInstance();
            factory.setNamespaceAware(true);
        }

        public Document parse(ComponentContext componentContext, InputSource inputSource) throws SAXException, IOException {

            errorMessages.clear();
            DocumentBuilder builder;
            try {
                builder = factory.newDocumentBuilder();
            } catch (ParserConfigurationException pe) {
                throw new SAXException(pe);
            }
            builder.setErrorHandler(this);
            builder.setEntityResolver(componentContext);
            Document document = builder.parse(inputSource);
            if (errorMessages.size() > 0) {
                Iterator<String> messageIterator = errorMessages.iterator();
                StringBuilder errorMessage = new StringBuilder();
                while (messageIterator.hasNext()) {
                    errorMessage.append(messageIterator.next() + "  ");
                }
                throw new SAXException(errorMessage.toString());
            }
            return document;
        }

        public void error(SAXParseException spe) throws SAXException {
            errorMessages.add(spe.getMessage());
        }

        public void warning(SAXParseException spe) throws SAXException {
            errorMessages.add(spe.getMessage());
        }

        public void fatalError(SAXParseException spe) throws SAXException {
            errorMessages.add(spe.getMessage());
        }

    }

}