/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.system;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import org.bluestemsoftware.specification.eoa.Dependency;
import org.bluestemsoftware.specification.eoa.DeploymentClassLoader;
import org.bluestemsoftware.specification.eoa.FactoryDependency;
import org.bluestemsoftware.specification.eoa.ProvidedDependency;
import org.bluestemsoftware.specification.eoa.ScopedDependency;
import org.bluestemsoftware.specification.eoa.SharedDependency;
import org.bluestemsoftware.specification.eoa.SystemDependencies;
import org.bluestemsoftware.specification.eoa.ext.ExtensionException;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactoryDeployment;

/**
 * Models a specific implementation, i.e. type, of dependency management system, e.g. Maven,
 * Ivy, etc ...
 * <p>
 * The type of dependency management system used by <code>System</code> is an extension to
 * the core specification.
 */
public abstract class DependencyManager {

    public static final String PROVIDER_CONFIGURATION_PATH = "META-INF/services/org.bluestemsoftware.specification.eoa.system.DependencyManager";

    /**
     * Invoked by <code>Container</code>. Allows <code>DependencyManager</code> to
     * initialize resources and create necessary state.
     * 
     * @param File
     *        configurationSettings used to configure <code>DependencyManager</code> or null
     *        if defaults are appropriate.
     * 
     * @throws ExtensionException
     */
    public abstract void init(File configurationSettings) throws ExtensionException;

    /**
     * Invoked by <code>Container</code> when <code>DependencyManager</code>. Allows
     * <code>DependencyManager</code> to clean-up resources and destroy state.
     */
    public abstract void destroy();

    /**
     * Returns a 'well-known' file name used by dependency management system to identify its
     * project descriptor document, e.g. 'pom.xml', 'ivy.xml', etc ...
     * <p>
     * Used by <code>Container</code> to locate type specific managed dependencies 'project'
     * descriptor document.
     * 
     * @return String well-known name
     */
    public abstract String getProjectDescriptorName();

    /**
     * Returns a 'well-known' file name used by dependency management system to identify its
     * system configuration file, e.g. 'settings.xml', ivyconf.xml, etc ...
     * <p>
     * This value MAY be used by <code>Container</code> to locate a type specific
     * configuration settings file.
     * 
     * @return String well-known name
     * @see #getContainerDependencies(org.w3c.dom.Document)
     */
    public abstract String getConfigurationFileName();

    /**
     * Generates 'managed' dependency model. Uses rules defined as an extension to this
     * specification which MUST by followed by <code>DependencyManager</code> when
     * discriminating between 'provided', 'shared', 'scoped' and 'factory' dependencies.
     * 
     * @param File
     *        project descriptor, i.e. a type specific document which models project
     *        dependencies.
     * 
     * @return SystemDependencies
     * @throws ExtensionException
     *         If an error occurs modeling dependencies.
     * 
     * @see ProvidedDependency
     * @see SharedDependency
     * @see ScopedDependency
     * @see FactoryDependency
     * @see ComponentDependency
     */
    public abstract SystemDependencies modelSystemDependencies(File projectDescriptor) throws ExtensionException;

    /**
     * 'Resolves' a dependency artifact from local repository. If not found,
     * <code>DependencyManager</code> will attempt to resolve the artifact from a remote
     * repository (defined within system project model). 
     * 
     * @param organizationID
     *        identifies the organization which authored dependency artifact
     * @param artifactID
     *        identifies dependency within context of organization
     * @param version
     *        dependency artifact version
     * @param extension
     *        dependency artifact extension, e.g. jar, eoa-factory, eoa-feature, eoa-component,
     *        etc ...
     * @return
     * @throws ExtensionException
     *         if an error occurs resolving dependency
     */
    public abstract Dependency resolveDependency(String organizationID, String artifactID, String version, String extension) throws ExtensionException;

    /**
     * Discovers and instantiates an implementation of <code>DependencyManager</code> using
     * java's 'service provider' loading mechanism which requires that a correctly formatted
     * service provider configuration file be located on system classpath and be named:
     * <p>
     * META-INF/services/org.bluestemsoftware.specification.eoa.system.DependencyManager
     * <p>
     * Note that while this requirement effectively defines the means by which an
     * implementation MAY be discovered and instantiated, the means by which impl and its
     * dependencies are deployed is <code>System</code> specific and is outside the scope of
     * this specification.
     * <p>
     * Note that this mechanism
     * @param SystemClassLoader
     * @return DependencyManager
     * @throws SystemException
     *         If an error occurs creating instance or if no implementation found.
     * 
     * @see java.util.ServiceLoader
     * @see ExtensionFactoryDeployment#getServiceProviderConfigurationFile(DeploymentClassLoader)
     */
    public static DependencyManager createInstance(SystemClassLoader systemClassLoader) throws SystemException {

        if (systemClassLoader == null) {
            throw new IllegalArgumentException("systemClassLoader null");
        }

        URL resource = systemClassLoader.findResource(PROVIDER_CONFIGURATION_PATH);
        if (resource == null) {
            throw new SystemException("Failed to locate a DependencyManager provider. Required service"
                    + " provider configuration file '"
                    + PROVIDER_CONFIGURATION_PATH
                    + "' not found by SystemClassLoader.");
        }

        List<String> classNames = new ArrayList<String>();
        BufferedReader reader = null;
        try {
            reader = new BufferedReader(new InputStreamReader(resource.openStream(), "UTF-8"));
            for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                if (line.indexOf('#') >= 0) {
                    line = line.substring(0, line.indexOf('#')).trim();
                }
                if (line.length() > 0) {
                    classNames.add(line);
                }
            }
        } catch (Exception ex) {
            throw new SystemException("Error parsing service provider configuration file located at '"
                    + resource.toString()
                    + "'. "
                    + ex);
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignore) {
                }
            }
        }

        if (classNames.size() != 1) {
            throw new SystemException("Service provider configuration file "
                    + " MUST specify exactly one implementation class.");
        }

        DependencyManager dependencyManager = null;
        try {
            Class<?> clazz = systemClassLoader.loadClass(classNames.get(0));
            dependencyManager = (DependencyManager)clazz.newInstance();
        } catch (Exception ex) {
            throw new SystemException("Error instantiating service provider. " + ex);
        }

        return dependencyManager;

    }

}
