/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.system.server;

import java.io.File;
import java.net.URI;

import org.bluestemsoftware.specification.eoa.ext.ManageableExtension;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureException;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureFactory;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.w3c.dom.Element;

/**
 * Encapsulates a specific type of abstract functionality as identified by
 * {@link #getExtensionType()}. Type specific methods are invoked by casting
 * <code>Feature</code> to a well-known implementation.
 * 
 * @see Server#getFeature(Class)
 */

public abstract class Feature extends ManageableExtension {

    protected String impl;
    protected File featureVarDir;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ManageableExtension.Provider {

        public String spi_getFeatureImpl();

        public void spi_setConfiguration(Element configuration);

        public void spi_init() throws FeatureException;

        public void spi_destroy();

    }

    public Feature(FeatureFactory factory, Provider provider) {
        super(factory, provider);
    }

    /**
     * Returns URI which uniquely identifies <code>Feature</code> implementation within
     * context of <code>Feature</code> type.
     * 
     * @return implementation URI
     */
    public String getFeatureImpl() {
        if (impl == null) {
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                impl = ((Provider)provider).spi_getFeatureImpl();
            } finally {
                thread.setContextClassLoader(cl);
            }
        }
        return impl;
    }

    /**
     * Used by <code>Server</code> to inject <code>Feature</code> configuration prior to
     * enabling feature.
     * <p>
     * Note that if no configuration is specified, a compliant feature MUST provide and utilize
     * a 'default' configuration.
     * 
     * @param configuration
     *        Root element of an impl specific configuration document.
     */
    public void setConfiguration(Element configuration) {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_setConfiguration(configuration);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Returns directory which MUST be utilized by <code>Feature</code> to store permanent
     * files of a variable nature, e.g. database files, queues, etc .... which MAY be required
     * to implement feature behavior.
     * <p>
     * To insure that directory is separate and apart from directory used by other instances of
     * <code>Feature</code>, the following convention is used to generate subdirectory name
     * from impl URI of <code>Feature</code>:
     * <p>
     * [scheme.] + [host.] + [port.] + [path]
     * <p>
     * where 'scheme', 'host', 'port' and 'path' are extracted from ref URI and all invalid
     * file name characters which MAY occur in [path] fragment are replaced with the '.'
     * character. As an example, if ref URI is:
     * <p>
     * org.bluestemsoftware.open.eoa.ext.feature.foo/myfeature
     * <p>
     * the file name is:
     * <p>
     * org.bluestemsoftware.open.eoa.ext.feature.foo.myfeature
     * 
     * @return File generated by resolving file name against base dir retrieved via
     *         {@link org.bluestemsoftware.specification.eoa.system.System#getSystemVarDir()}.
     */
    public File getFeatureVarDir() {
        if (featureVarDir == null) {
            if (getFeatureImpl().length() == 0) {
                throw new IllegalStateException("Feature impl URI is empty string");
            }
            try {
                URI uri = new URI(getFeatureImpl());
                String location = uri.getScheme() == null ? "" : uri.getScheme() + '.';
                location = uri.getHost() == null ? location : location + '.' + uri.getHost();
                location = uri.getPort() < 0 ? location : location + '.' + uri.getPort();
                String path = uri.getPath();
                if (path != null) {
                    path = path.replaceAll("/", ".");
                    path = path.replaceAll("\\\\", ".");
                    path = path.replaceAll(":", ".");
                    path = path.replaceAll("\\*", ".");
                    path = path.replaceAll("\\?", ".");
                    path = path.replaceAll("\"", ".");
                    path = path.replaceAll("<", ".");
                    path = path.replaceAll(">", ".");
                    path = path.replaceAll("\\|", ".");
                }
                if (location.length() > 0) {
                    location = path == null ? location : location + '.' + path;
                } else {
                    location = path;
                }
                File systemVarDir = SystemContext.getContext().getSystem().getSystemVarDir();
                featureVarDir = new File(systemVarDir, location);
                if (!featureVarDir.exists()) {
                    featureVarDir.mkdir();
                }
            } catch (Exception ex) {
                throw new IllegalStateException("Error creating featureVarDir. " + ex);
            }
        }
        return featureVarDir;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ManageableExtension#init()
     */
    public void init() throws FeatureException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_init();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ManageableExtension#destroy()
     */
    public void destroy() {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_destroy();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("type: ");
        sb.append(getExtensionType());
        sb.append(System.getProperty("line.separator"));
        sb.append("impl: ");
        sb.append(getFeatureImpl());
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(getExtensionType()) + hashCode(getFeatureImpl());
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Feature)) {
            return false;
        }

        if ((getExtensionType() == ((Feature)obj).getExtensionType() || (getExtensionType() != null && getExtensionType()
                .equals(((Feature)obj).getExtensionType())))
                && (getFeatureImpl() == ((Feature)obj).getFeatureImpl() || (getFeatureImpl() != null && getFeatureImpl()
                        .equals(((Feature)obj).getFeatureImpl())))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}