/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.system.server;

import java.util.List;

import org.bluestemsoftware.specification.eoa.FeatureDependency;
import org.bluestemsoftware.specification.eoa.SystemDependencies;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactory;
import org.bluestemsoftware.specification.eoa.ext.ManageableExtension;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.rt.TransportProtocolRT;
import org.w3c.dom.Element;

/**
 * Encapsulates 'server' functionality implemented by <code>System</code>. Provides a
 * configurable set of features which MAY be required by extensions to implement their
 * behavior.
 */
public abstract class Server extends ManageableExtension {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ManageableExtension.Provider {

    }

    public Server(ExtensionFactory factory, Provider provider) {
        super(factory, provider);
    }

    /**
     * Retrieves a <code>Feature</code> implemented by <code>Server</code> by
     * <code>Feature</code> class. Features encapsulate a specific type of abstract
     * functionality as identified by feature type.
     * <p>
     * Note that if more than one implementation of a specific type of feature factory is
     * defined on <code>System</code>, <code>Server</code> MUST, provide a means by which
     * a specific impl and configuration MAY be indicated by user. If not otherwise indicated,
     * the implementation chosen is arbitrary.
     * 
     * @param featureType
     *        identifies type of <code>Feature</code> to retrieve
     * 
     * @return A <code>Feature</code> component of indicated type or <code>null</code> if
     *         <code>Feature</code> is undefined.
     * 
     * @see Feature#getExtensionType()
     * @see FeatureDependency#isRequired()
     * @see SystemDependencies#getFactoryDependencies()
     */
    public abstract <T extends Feature> T getFeature(Class<T> featureType);

    /**
     * Retrieves a <code>Feature</code> implemented by <code>Server</code> by type ref URI.
     * Features encapsulate a specific type of abstract functionality as identified by feature
     * type.
     * <p>
     * Note that if more than one implementation of a specific type of feature factory is
     * defined on <code>System</code>, <code>Server</code> MUST, provide a means by which
     * a specific impl and configuration MAY be indicated by user, i.e. 'managed'. If the
     * feature is un-managed, the implementation chosen is arbitrary.
     * 
     * @param featureType
     *        type ref URI of <code>Feature</code> to retrieve
     * 
     * @return A <code>Feature</code> component of indicated type or <code>null</code> if
     *         <code>Feature</code> is undefined.
     * 
     * @see Feature#getExtensionType()
     * @see FeatureDependency#isRequired()
     * @see SystemDependencies#getFactoryDependencies()
     */
    public abstract Feature getFeature(String featureType);

    /**
     * Retrieves all features implemented by <code>Server</code> which MUST be listed in the
     * order in which they were enabled, i.e. instantiated and initialized.
     * <p>
     * Note that before a <code>Feature</code> is enabled, a compliant server MUST ensure
     * that all required, dependent features have been enabled.
     * <p>
     * Note that a compliant <code>Server</code> must disable, i.e. destroy, features in the
     * reverse order in which they were enabled.
     * 
     * 
     * @return List of features or an empty List if none implemented.
     */
    public abstract List<Feature> getFeatures();

    /**
     * Gets <code>TransportFeature</code> by transport type.
     * 
     * @param transportType
     * @return transport feature or <code>null</code> if undefined.
     * 
     */
    public abstract TransportFeature getTransportFeature(Class<? extends TransportProtocolRT> transportType);

    /**
     * Invoked by <code>Feature</code> when configuration changes. A compliant
     * <code>Server</code> must persist the supplied configuration and use it to enable
     * <code>Feature</code> when server is re-started.
     * 
     * @param feature
     *        <code>Feature</code> for which configuration is applicable.
     * 
     * @param configuration
     * @see Feature#setConfiguration(Element)
     */
    public abstract void updateFeatureConfiguration(Feature feature, Element configuration);

}